#! /usr/bin/env python
#
# Connect to NETCONF server using unix socket passed on the command line.
# Then display the servers capabilities. After connecting you can also use 
# all other operations supported by ncclient, example: get(), get_config()...
#
# To test against a example NETCONF server with unix socket support check out:
# https://github.com/CESNET/libnetconf2
# It contains an example server which can listen for connections using a unix 
# socket.
#
# $ ./nc10.py "/path/to/socket"

import sys, warnings
warnings.simplefilter("ignore", DeprecationWarning)
from ncclient import manager

def demo(path):
    with manager.connect_uds(path) as m:
        for c in m.server_capabilities:
            print(c)

if __name__ == '__main__':
    demo(sys.argv[1])


import unittest
import os
import sys

from ncclient import manager
from ncclient.xml_ import *
file_path = os.path.join(os.getcwd(), "test", "unit", "reply1")


class Test_NCElement(unittest.TestCase):

    def test_ncelement_reply_001(self):
        """test parse rpc_reply and string/data_xml"""
        # read in reply1 contents
        with open(file_path, 'r') as f:
            reply = f.read()
        device_params = {'name': 'junos'}
        device_handler = manager.make_device_handler(device_params)
        transform_reply = device_handler.transform_reply()
        result = NCElement(reply, transform_reply)
        result_str = result.tostring
        result_str = result_str.decode('UTF-8')
        self.assertEqual(str(result), result_str)
        #data_xml != tostring
        self.assertNotEqual(result_str, result.data_xml)

    def test_ncelement_reply_002(self):
        """test parse rpc_reply and xpath"""
        # read in reply1 contents
        with open(file_path, 'r') as f:
            reply = f.read()
        device_params = {'name': 'junos'}
        device_handler = manager.make_device_handler(device_params)
        transform_reply = device_handler.transform_reply()
        result = NCElement(reply, transform_reply)
        # XPATH checks work
        self.assertEqual(result.xpath("//host-name")[0].text, "R1")
        self.assertEqual(
            result.xpath("/rpc-reply/software-information/host-name")[0].text,
            "R1")
        self.assertEqual(
            result.xpath("software-information/host-name")[0].text,
            "R1")

    def test_ncelement_reply_003(self):
        """test parse rpc_reply and find"""
        # read in reply1 contents
        with open(file_path, 'r') as f:
            reply = f.read()
        device_params = {'name': 'junos'}
        device_handler = manager.make_device_handler(device_params)
        transform_reply = device_handler.transform_reply()
        result = NCElement(reply, transform_reply)
        # find
        self.assertEqual(result.findtext(".//host-name"), "R1")
        self.assertEqual(result.find(".//host-name").tag, "host-name")
        self.assertEqual(result.find(".//host-name"), result.findall(".//host-name")[0])
        self.assertEqual(result.findall(".//host-name")[0].tag, "host-name")


class TestXML(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.f = open(file_path, 'r')
        cls.reply = cls.f.read()

    def test_ncelement_reply_001(self):
        device_params = {'name': 'junos'}
        device_handler = manager.make_device_handler(device_params)
        transform_reply = device_handler.transform_reply()
        result = NCElement(self.reply, transform_reply)
        self.assertEqual(result.xpath("//name")[0].text, "junos")
        self.assertEqual(result.xpath("//name")[0].tag, "name")
        self.assertEqual(
            result.xpath("//package-information")[0].tag,
            "package-information")

    def test_ncelement_find(self):
        device_params = {'name': 'junos'}
        device_handler = manager.make_device_handler(device_params)
        transform_reply = device_handler.transform_reply()
        result = NCElement(self.reply, transform_reply)
        self.assertEqual(result.find(".//name").tag, "name")
        self.assertEqual(result.find(".//name").text, "junos")

    def test_ncelement_findtext(self):
        device_params = {'name': 'junos'}
        device_handler = manager.make_device_handler(device_params)
        transform_reply = device_handler.transform_reply()
        result = NCElement(self.reply, transform_reply)
        self.assertEqual(result.findtext(".//name"), "junos")

    def test_ncelement_findall(self):
        device_params = {'name': 'junos'}
        device_handler = manager.make_device_handler(device_params)
        transform_reply = device_handler.transform_reply()
        result = NCElement(self.reply, transform_reply)
        self.assertEqual(result.findall(".//name")[0].tag, "name")
        self.assertEqual(result.findall(".//name")[0].text, "junos")

    def test_ncelement_remove_namespace(self):
        device_params = {'name': 'junos'}
        device_handler = manager.make_device_handler(device_params)
        transform_reply = device_handler.transform_reply()
        result = NCElement(self.reply, transform_reply)
        re = result.remove_namespaces(self.reply)
        self.assertEqual(re.tag, "rpc-reply")
        ele = to_ele((result.find(".//name")))
        self.assertEqual(ele.tag, "name")
        self.assertEqual(ele.text, "junos")
        ele = to_ele(self.reply)
        self.assertEqual(ele.tag, "rpc-reply")

    def test_to_ele(self):
        device_params = {'name': 'junos'}
        device_handler = manager.make_device_handler(device_params)
        transform_reply = device_handler.transform_reply()
        result = NCElement(self.reply, transform_reply)
        ele = to_ele((result.find(".//name")))
        self.assertEqual(ele.tag, "name")
        self.assertEqual(ele.text, "junos")
        ele = to_ele(self.reply)
        self.assertEqual(ele.tag, "rpc-reply")

    def test_parse_root(self):
        device_params = {'name': 'junos'}
        device_handler = manager.make_device_handler(device_params)
        transform_reply = device_handler.transform_reply()
        result = NCElement(self.reply, transform_reply)
        tag, attrib = parse_root(result.data_xml)
        self.assertEqual(tag, "rpc-reply")
        self.assertEqual(attrib, {'attrib1': 'test'})

    def test_validated_element_pass(self):
        device_params = {'name': 'junos'}
        device_handler = manager.make_device_handler(device_params)
        transform_reply = device_handler.transform_reply()
        result = NCElement(self.reply, transform_reply)
        result_xml = result.data_xml
        ele = validated_element(
            result_xml, tags=["rpc-reply", "rpc"], attrs=[["attrib1", "attrib2"]])
        self.assertEqual(ele.tag, "rpc-reply")
        ele = validated_element(
            result_xml, attrs=[["attrib1", "attrib2"]])
        self.assertEqual(ele.tag, "rpc-reply")
        ele = validated_element(result_xml)
        self.assertEqual(ele.tag, "rpc-reply")

    def test_validated_element_fail(self):
        device_params = {'name': 'junos'}
        device_handler = manager.make_device_handler(device_params)
        transform_reply = device_handler.transform_reply()
        result = NCElement(self.reply, transform_reply)
        XMLError.message = "Element does not meet requirement"
        result_xml = result.data_xml
        self.assertRaises(XMLError,
            validated_element,
                result_xml, tags=["rpc"], attrs=[["attrib1", "attrib2"]])

    def test_validated_element_fail_2(self):
        device_params = {'name': 'junos'}
        device_handler = manager.make_device_handler(device_params)
        transform_reply = device_handler.transform_reply()
        result = NCElement(self.reply, transform_reply)
        XMLError.message = "Element does not meet requirement"
        result_xml = result.data_xml
        self.assertRaises(XMLError,
            validated_element,
                result_xml,
                tags=[
                    "rpc-reply",
                    "rpc"],
                attrs=[
                    ["attrib1"],
                    ["attrib2"]])

    def test_validated_element_fail_3(self):
        device_params = {'name': 'junos'}
        device_handler = manager.make_device_handler(device_params)
        transform_reply = device_handler.transform_reply()
        result = NCElement(self.reply, transform_reply)
        XMLError.message = "Element does not meet requirement"
        result_xml = result.data_xml
        self.assertRaises(XMLError,
            validated_element, result_xml, tags=["rpc"])

    def test_sub_ele_inherit_parent_namespace(self):
        device_params = {'name': 'junos'}
        device_handler = manager.make_device_handler(device_params)
        transform_reply = device_handler.transform_reply()
        result = NCElement(self.reply, transform_reply)
        ele = new_ele_ns(result.find("./cli").tag, "http://www.xxx.org")
        child = sub_ele(ele, "child")
        sibling = sub_ele(ele, "sibling")
        grandchild = sub_ele(child, "grandchild")
        self.assertEqual(child.tag, "{http://www.xxx.org}child")
        self.assertEqual(sibling.tag, "{http://www.xxx.org}sibling")
        self.assertEqual(grandchild.tag, "{http://www.xxx.org}grandchild")



import unittest
from ncclient.devices.hpcomware import *
from ncclient.xml_ import *


class TestHpcomwareDevice(unittest.TestCase):

    def setUp(self):
        self.obj = HpcomwareDeviceHandler({'name': 'hpcomware'})

    def test_get_xml_base_namespace_dict(self):
        expected = {None: BASE_NS_1_0}
        self.assertDictEqual(expected, self.obj.get_xml_base_namespace_dict())

    def test_get_xml_extra_prefix_kwargs(self):
        expected = {'nsmap': {'data': 'http://www.hp.com/netconf/data:1.0',
                              'config': 'http://www.hp.com/netconf/config:1.0',
                              None: 'urn:ietf:params:xml:ns:netconf:base:1.0'}}
        self.assertDictEqual(expected, self.obj.get_xml_extra_prefix_kwargs())

    def test_add_additional_operations(self):
        expected = dict()
        expected['cli_display'] = DisplayCommand
        expected['cli_config'] = ConfigCommand
        expected['action'] = Action
        expected['save'] = Save
        expected['rollback'] = Rollback
        self.assertDictEqual(expected, self.obj.add_additional_operations())



#!/usr/bin/env python
#

import logging
import sys

from ncclient import manager
from ncclient.xml_ import to_xml
from ncclient.operations.rpc import RPCError

_NS_MAP = {
    'nc': 'urn:ietf:params:xml:ns:netconf:base:1.0',
    'nokia-conf': 'urn:nokia.com:sros:ns:yang:sr:conf'
}

_NOKIA_MGMT_INT_FILTER = '''
<configure xmlns="%s">
  <system>
    <management-interface/>
  </system>
</configure>
''' % (_NS_MAP['nokia-conf'])

def connect(host, port, user, password):
    m = manager.connect(host=host, port=port,
            username=user, password=password,
            device_params={'name': 'sros'},
            hostkey_verify=False)

    ## Retrieve full configuration from the running datastore
    running_xml = m.get_config(source='running')
    logging.info(running_xml)

    ## Retrieve full configuration from the running datastore and strip
    ## the rpc-reply + data elements
    running_xml = m.get_config(source='running').xpath(
            '/nc:rpc-reply/nc:data/nokia-conf:configure',
            namespaces=_NS_MAP)[0]
    logging.info(to_xml(running_xml, pretty_print=True))

    ## Retrieve full configuration from the running datastore and strip
    ## out elements except for /configure/system/management-interface
    running_xml = m.get_config(source='running').xpath(
            '/nc:rpc-reply/nc:data/nokia-conf:configure' \
            '/nokia-conf:system/nokia-conf:management-interface',
            namespaces=_NS_MAP)[0]
    logging.info(to_xml(running_xml, pretty_print=True))

    ## Retrieve a specific filtered subtree from the running datastore
    ## and handle any rpc-error should a portion of the filter criteria
    ## be invalid
    try:
        running_xml = m.get_config(source='running',
                filter=('subtree', _NOKIA_MGMT_INT_FILTER),
                with_defaults='report-all').xpath(
                        '/nc:rpc-reply/nc:data/nokia-conf:configure',
                        namespaces=_NS_MAP)[0]
        logging.info(to_xml(running_xml, pretty_print=True))

    except RPCError as err:
        logging.info('Error: %s' % err.message.strip())

    m.close_session()

if __name__ == '__main__':
    LOG_FORMAT = '%(asctime)s %(levelname)s %(filename)s:%(lineno)d %(message)s'
    logging.basicConfig(stream=sys.stdout, level=logging.INFO, format=LOG_FORMAT)

    try:
        connect(sys.argv[1], '830', 'admin', 'admin')

    except IndexError:
        logging.error('Must supply a valid hostname or IP address')




import unittest
from ncclient.devices.default import DefaultDeviceHandler


capabilities = ['urn:ietf:params:netconf:base:1.0',
                'urn:ietf:params:netconf:base:1.1',
                'urn:ietf:params:netconf:capability:writable-running:1.0',
                'urn:ietf:params:netconf:capability:candidate:1.0',
                'urn:ietf:params:netconf:capability:confirmed-commit:1.0',
                'urn:ietf:params:netconf:capability:rollback-on-error:1.0',
                'urn:ietf:params:netconf:capability:startup:1.0',
                'urn:ietf:params:netconf:capability:url:1.0?scheme=http,ftp,file,https,sftp',
                'urn:ietf:params:netconf:capability:validate:1.0',
                'urn:ietf:params:netconf:capability:xpath:1.0',
                'urn:ietf:params:netconf:capability:notification:1.0',
                'urn:ietf:params:netconf:capability:interleave:1.0',
                'urn:ietf:params:netconf:capability:with-defaults:1.0']

class TestDefaultDevice(unittest.TestCase):

    def setUp(self):
        self.obj = DefaultDeviceHandler()

    def test_get_capabilties(self):
        self.assertEqual(self.obj.get_capabilities(), capabilities)

    def test_get_ssh_subsystem_names(self):
        self.assertEqual(self.obj.get_ssh_subsystem_names(), ["netconf"])

    def test_perform_qualify_check(self):
        self.assertTrue(self.obj.perform_qualify_check())

    def test_handle_raw_dispatch(self):
        self.assertFalse(self.obj.handle_raw_dispatch(None))

    def test_handle_connection_exceptions(self):
        self.assertFalse(self.obj.handle_connection_exceptions(None))

    def test_is_rpc_error_exempt_1(self):
        self.assertFalse(self.obj.is_rpc_error_exempt(None))

    def test_is_rpc_error_exempt_2(self):
        self.obj._exempt_errors_exact_match = ["test_exempt"]
        self.assertTrue(self.obj.is_rpc_error_exempt("  Test_Exempt"))

    def test_is_rpc_error_exempt_3(self):
        self.obj._exempt_errors_startwith_wildcard_match = ["test_exempt"]
        self.assertTrue(self.obj.is_rpc_error_exempt("*Test_Exempt"))

    def test_is_rpc_error_exempt_4(self):
        self.obj._exempt_errors_endwith_wildcard_match = ["test_exempt"]
        self.assertTrue(self.obj.is_rpc_error_exempt("Test_Exempt*"))

    def test_is_rpc_error_exempt_5(self):
        self.obj._exempt_errors_full_wildcard_match = ["test_exempt"]
        self.assertTrue(self.obj.is_rpc_error_exempt("*Test_Exempt*"))

suite = unittest.TestSuite()
unittest.TextTestRunner().run(suite)




import sys
import unittest
import socket
from ncclient.devices.junos import JunosDeviceHandler

try:
    from unittest.mock import MagicMock, patch, call
except ImportError:
    from mock import MagicMock, patch, call

# only import on non-Windows platforms
if sys.platform != 'win32':
    from ncclient.transport.errors import UnixSocketError
    from ncclient.transport.unixSocket import UnixSocketSession

PATH = '/tmp/test_socket.sock'

class TestUnixSocket(unittest.TestCase):

    @unittest.skipIf(sys.platform.startswith('win'), "Skipping on Windows")
    @patch('socket.socket.close')
    def test_close_UnixSocket(self, mock_sock_close_fn):
        session = UnixSocketSession(MagicMock())
        session._socket = socket.socket()
        session._connected = True
        session.close()
        mock_sock_close_fn.assert_called_once_with()
        self.assertFalse(session.connected)

    @unittest.skipIf(sys.platform.startswith('win'), "Skipping on Windows")
    @patch('socket.socket')
    @patch('ncclient.transport.Session._post_connect')
    def test_connect_UnixSocket(self, mock_post_connect, mock_socket):
        session = UnixSocketSession(MagicMock())
        session.connect(path=PATH)
        self.assertTrue(session.connected)



import logging
import sys

from ncclient import manager

# Global constants
DEVICE_HOST = 'localhost'
DEVICE_PORT = 830
DEVICE_USER = 'admin'
DEVICE_PASS = 'admin'

EDIT_CONFIG_PAYLOAD = """
<config xmlns="urn:ietf:params:xml:ns:netconf:base:1.0" xmlns:alu="urn:ietf:params:xml:ns:netconf:base:1.0">
    <configure xmlns="urn:nokia.com:sros:ns:yang:sr:conf">
        <system>
            <security>
                <snmp>
                    <community alu:operation="replace">
                        <community-string>Sample Community</community-string>
                        <access-permissions>r</access-permissions>
                        <version>v2c</version>
                    </community>
                </snmp>
            </security>
        </system>
    </configure>
</config>
"""


def create_session(host, port, username, password):
    """Creates and returns an ncclient manager session."""
    return manager.connect(
        host=host,
        port=port,
        username=username,
        password=password,
        hostkey_verify=False,
        device_params={'name': 'sros'}
    )


def edit_config(m, config_payload):
    """Edits the configuration with the given payload."""
    m.edit_config(target='candidate', config=config_payload)


def commit_changes(m, comment=''):
    """Commits changes with an optional comment."""
    response = m.commit(comment=comment)
    logging.info(response)


def main():
    LOG_FORMAT = '%(asctime)s %(levelname)s %(filename)s:%(lineno)d %(message)s'
    logging.basicConfig(stream=sys.stdout, level=logging.INFO, format=LOG_FORMAT)

    try:
        with create_session(DEVICE_HOST, DEVICE_PORT, DEVICE_USER, DEVICE_PASS) as m:
            # Lock the configuration
            m.lock(target='candidate')

            # Edit the configuration
            edit_config(m, EDIT_CONFIG_PAYLOAD)

            # Commit the configuration
            commit_changes(m, comment='A sample commit comment goes here')

            # Unlock the configuration
            m.unlock(target='candidate')

    except Exception as e:
        logging.error(f"Encountered an error: {e}")


if __name__ == '__main__':
    main()



from ncclient.operations.edit import *
from ncclient.operations.subscribe import *
import unittest
try:
    from unittest.mock import patch, MagicMock  # Python 3.4 and later
except ImportError:
    from mock import patch, MagicMock
from ncclient import manager
import ncclient.manager
import ncclient.transport
from ncclient.xml_ import *
from ncclient.operations import RaiseMode
from xml.etree import ElementTree
from ncclient.operations.errors import MissingCapabilityError
import copy

start_time = "1990-12-31T23:59:60Z"
stop_time = "1996-12-19T16:39:57-08:00"

class TestSubscribe(unittest.TestCase):

    def setUp(self):
        self.device_handler = manager.make_device_handler({'name': 'junos'})

    @patch('ncclient.operations.edit.RPC._request')
    def test_subscribe_all(self, mock_request):
        session = ncclient.transport.SSHSession(self.device_handler)
        session._server_capabilities = [":notification"]
        obj = CreateSubscription(
            session,
            self.device_handler,
            raise_mode=RaiseMode.ALL)
        obj.request()
        node = new_ele_ns("create-subscription", NETCONF_NOTIFICATION_NS)
        xml = ElementTree.tostring(node)
        call = mock_request.call_args_list[0][0][0]
        call = ElementTree.tostring(call)
        self.assertEqual(call, xml)

    @patch('ncclient.operations.edit.RPC._request')
    def test_subscribe_stream(self, mock_request):
        session = ncclient.transport.SSHSession(self.device_handler)
        session._server_capabilities = [":notification"]
        obj = CreateSubscription(
            session,
            self.device_handler,
            raise_mode=RaiseMode.ALL)
        obj.request(filter=None, stream_name="nameofstream")
        node = new_ele_ns("create-subscription", NETCONF_NOTIFICATION_NS)
        sub_ele_ns(node, "stream", NETCONF_NOTIFICATION_NS).text = "nameofstream"
        xml = ElementTree.tostring(node)
        call = mock_request.call_args_list[0][0][0]
        call = ElementTree.tostring(call)
        self.assertEqual(call, xml)

    @patch('ncclient.operations.edit.RPC._request')
    def test_subscribe_times(self, mock_request):
        session = ncclient.transport.SSHSession(self.device_handler)
        session._server_capabilities = [":notification"]
        obj = CreateSubscription(
            session,
            self.device_handler,
            raise_mode=RaiseMode.ALL)
        obj.request(filter=None, start_time=start_time, stop_time=stop_time)
        node = new_ele_ns("create-subscription", NETCONF_NOTIFICATION_NS)
        sub_ele_ns(node, "startTime", NETCONF_NOTIFICATION_NS).text = start_time
        sub_ele_ns(node, "stopTime", NETCONF_NOTIFICATION_NS).text = stop_time
        xml = ElementTree.tostring(node)
        call = mock_request.call_args_list[0][0][0]
        call = ElementTree.tostring(call)
        self.assertEqual(call, xml)



#! /usr/bin/env python
#
# Configure an Interface: its description and make it active.
# XML payload created with lxml/etree instead of a template
# 
# $ ./nc08.py Paulo Seguel

import sys, warnings
warnings.simplefilter("ignore", DeprecationWarning)
import datetime
from ncclient import manager
from lxml import etree

current_time = datetime.datetime.strftime(datetime.datetime.now(), '%Y-%m-%d %H:%M:%S')
 
# build xml
config_e = etree.Element("config")
configuration = etree.SubElement(config_e, "interface-configurations", nsmap = {None: 'http://cisco.com/ns/yang/Cisco-IOS-XR-ifmgr-cfg'})
interface_cfg = etree.SubElement(configuration, "interface-configuration")
active = etree.SubElement(interface_cfg, "active").text = 'act'
interface_name = etree.SubElement(interface_cfg, "interface-name").text = 'GigabitEthernet0/0/0/0'
description = etree.SubElement(interface_cfg, "description").text  = 'NETCONF configured - ' + current_time

def demo(host, user, password):
    with manager.connect(host=host, port=830, username=user, password=password,
                     hostkey_verify=False, device_params={'name':'default'},
                     look_for_keys=False, allow_agent=False) as m:
        with m.locked(target="candidate"):
            m.edit_config(config=config_e, default_operation="merge", target="candidate")
            m.commit()

if __name__ == '__main__':
    demo(sys.argv[1], sys.argv[2], sys.argv[3])




#!/usr/bin/env python
# Python script to fetch interface name and their operation status
import sys

import logging

from ncclient import manager


def connect(host, port, user, password):
    conn = manager.connect(host=host,
                           port=port,
                           username=user,
                           password=password,
                           timeout=60,
                           device_params={'name': 'junos'},
                           hostkey_verify=False)

    rpc = "<get-interface-information><terse/></get-interface-information>"
    response = conn.rpc(rpc)
    interface_name = response.xpath('//physical-interface/name')
    interface_status = response.xpath('//physical-interface/oper-status')
    for name, status in zip(interface_name, interface_status):
        name = name.text.split('\n')[1]
        status = status.text.split('\n')[1]
        logging.info("{} - {}".format(name, status))


if __name__ == '__main__':
    LOG_FORMAT = '%(asctime)s %(levelname)s %(filename)s:%(lineno)d %(message)s'
    logging.basicConfig(stream=sys.stdout, level=logging.INFO, format=LOG_FORMAT)

    connect('router', 830, 'netconf', 'juniper!')




"""
Handler for Juniper device specific information.

Note that for proper import, the classname has to be:

    "<Devicename>DeviceHandler"

...where <Devicename> is something like "Default", "Junos", etc.

All device-specific handlers derive from the DefaultDeviceHandler, which implements the
generic information needed for interaction with a Netconf server.

"""
import logging
import re

from lxml import etree
from lxml.etree import QName

from ncclient.operations.retrieve import GetSchemaReply
from .default import DefaultDeviceHandler
from ncclient.operations.third_party.juniper.rpc import GetConfiguration, LoadConfiguration, CompareConfiguration
from ncclient.operations.third_party.juniper.rpc import ExecuteRpc, Command, Reboot, Halt, Commit, Rollback
from ncclient.operations.rpc import RPCError
from ncclient.xml_ import to_ele, replace_namespace, BASE_NS_1_0, NETCONF_MONITORING_NS
from ncclient.transport.third_party.junos.parser import JunosXMLParser
from ncclient.transport.parser import DefaultXMLParser
from ncclient.transport.parser import SAXParserHandler


logger = logging.getLogger(__name__)


class JunosDeviceHandler(DefaultDeviceHandler):
    """
    Juniper handler for device specific information.

    """

    def __init__(self, device_params, ignore_errors=None):
        super(JunosDeviceHandler, self).__init__(device_params, ignore_errors)
        self.__reply_parsing_error_transform_by_cls = {
            GetSchemaReply: fix_get_schema_reply
        }

    def add_additional_operations(self):
        dict = {}
        dict["rpc"] = ExecuteRpc
        dict["get_configuration"] = GetConfiguration
        dict["load_configuration"] = LoadConfiguration
        dict["compare_configuration"] = CompareConfiguration
        dict["command"] = Command
        dict["reboot"] = Reboot
        dict["halt"] = Halt
        dict["commit"] = Commit
        dict["rollback"] = Rollback
        return dict

    def perform_qualify_check(self):
        return False

    def handle_raw_dispatch(self, raw):
        if 'routing-engine' in raw:
            raw = re.sub(r'<ok/>', '</routing-engine>\n<ok/>', raw)
            return raw
        # check if error is during capabilities exchange itself
        elif re.search(r'<rpc-reply>.*?</rpc-reply>.*</hello>?', raw, re.M | re.S):
            errs = re.findall(
                r'<rpc-error>.*?</rpc-error>', raw, re.M | re.S)
            err_list = []
            if errs:
                add_ns = """
                        <xsl:stylesheet version="1.0" xmlns:xsl="http://www.w3.org/1999/XSL/Transform">
                          <xsl:output indent="yes"/>
                            <xsl:template match="*">
                            <xsl:element name="{local-name()}" namespace="urn:ietf:params:xml:ns:netconf:base:1.0">
                            <xsl:apply-templates select="@*|node()"/>
                            </xsl:element>
                          </xsl:template>
                        </xsl:stylesheet>"""
                for err in errs:
                    doc = etree.ElementTree(etree.XML(err))
                    # Adding namespace using xslt
                    xslt = etree.XSLT(etree.XML(add_ns))
                    transformed_xml = etree.XML(etree.tostring(xslt(doc)))
                    err_list.append(RPCError(transformed_xml))
                return RPCError(to_ele("<rpc-reply>"+''.join(errs)+"</rpc-reply>"), err_list)
        else:
            return False

    def handle_connection_exceptions(self, sshsession):
        c = sshsession._channel = sshsession._transport.open_channel(
            kind="session")
        c.set_name("netconf-command-" + str(sshsession._channel_id))
        c.exec_command("xml-mode netconf need-trailer")
        return True

    def reply_parsing_error_transform(self, reply_cls):
        # return transform function if found, else None
        return self.__reply_parsing_error_transform_by_cls.get(reply_cls)

    def transform_reply(self):
        reply = '''<xsl:stylesheet version="1.0" xmlns:xsl="http://www.w3.org/1999/XSL/Transform">
        <xsl:output method="xml" indent="no"/>

        <xsl:template match="/|comment()|processing-instruction()">
            <xsl:copy>
                <xsl:apply-templates/>
            </xsl:copy>
        </xsl:template>

        <xsl:template match="*">
            <xsl:element name="{local-name()}">
                <xsl:apply-templates select="@*|node()"/>
            </xsl:element>
        </xsl:template>

        <xsl:template match="@*">
            <xsl:attribute name="{local-name()}">
                <xsl:value-of select="."/>
            </xsl:attribute>
        </xsl:template>
        </xsl:stylesheet>
        '''
        return reply.encode('UTF-8')

    def get_xml_parser(self, session):
        # use_filter in device_params can be used to enabled using SAX parsing
        if self.device_params.get('use_filter', False):
            l = session.get_listener_instance(SAXParserHandler)
            if l:
                session.remove_listener(l)
                del l
            session.add_listener(SAXParserHandler(session))
            return JunosXMLParser(session)
        else:
            return DefaultXMLParser(session)


def fix_get_schema_reply(root):
    # Workaround for wrong namespace of the data elem
    # (issue with some Junos versions, might be corrected by Juniper at some point)

    # get the data element, by local-name
    data_elems = root.xpath('/nc:rpc-reply/*[local-name()="data"]', namespaces={'nc': BASE_NS_1_0})

    if len(data_elems) != 1:
        return  # Will not alter unexpected content

    data_el = data_elems[0]
    namespace = QName(data_el).namespace

    if namespace == BASE_NS_1_0:
        # With the default netconf setting, we may get "{BASE_NS_1_0}data"; warn and fix it
        logger.warning("The device seems to run non-rfc compliant netconf. You may want to "
                       "configure: 'set system services netconf rfc-compliant'")
        replace_namespace(data_el, old_ns=BASE_NS_1_0, new_ns=NETCONF_MONITORING_NS)
    elif namespace is None:
        # With 'set system services netconf rfc-compliant' we may get "data" (no namespace); fix it
        # There is no default xmlns and the data el is <data xmlns:ncm="NETCONF_MONITORING_NS">
        replace_namespace(data_el, old_ns=None, new_ns=NETCONF_MONITORING_NS)



"""
Handler for Cisco CSR device specific information.

Note that for proper import, the classname has to be:

    "<Devicename>DeviceHandler"

...where <Devicename> is something like "Default", "Nexus", etc.

All device-specific handlers derive from the DefaultDeviceHandler, which implements the
generic information needed for interaction with a Netconf server.

"""


from .default import DefaultDeviceHandler
from warnings import warn

def csr_unknown_host_cb(host, fingerprint):
        #This will ignore the unknown host check when connecting to CSR devices
        return True

class CsrDeviceHandler(DefaultDeviceHandler):
    """
    Cisco CSR handler for device specific information.

    """
    def __init__(self, device_params, ignore_errors=None):
        warn(
            'CsrDeviceHandler is deprecated, please use IosxeDeviceHandler',
            DeprecationWarning,
            stacklevel=2)
        super(CsrDeviceHandler, self).__init__(device_params, ignore_errors)

    def add_additional_ssh_connect_params(self, kwargs):
        warn(
            'CsrDeviceHandler is deprecated, please use IosxeDeviceHandler',
            DeprecationWarning,
            stacklevel=2)
        kwargs['unknown_host_cb'] = csr_unknown_host_cb



import unittest
from ncclient import manager
from ncclient.devices.ericsson import *
from ncclient.operations.rpc import *
from ncclient.capabilities import Capabilities
import ncclient.transport

class TestEricssonDevice(unittest.TestCase):

    def setUp(self):
        self.device_handler = manager.make_device_handler({'name': 'ericsson'})

    def test_rpc_default(self):
        # It is a switch for user to turn on/off "nc" prefix, the "nc" prefix is disable by default
        session = ncclient.transport.SSHSession(self.device_handler)
        obj = RPC(session, self.device_handler, raise_mode=RaiseMode.ALL, timeout=0)

        expected = """<?xml version="1.0" encoding="UTF-8"?><rpc xmlns="urn:ietf:params:xml:ns:netconf:base:1.0" message-id="%s"><get-config><source><running/></source></get-config></rpc>""" % obj.id

        node = new_ele("get-config")
        child = sub_ele(node, "source")
        sub_ele(child, "running")

        rpc_node = obj._wrap(node)
        self.assertEqual(rpc_node, expected)

    def test_rpc_disable_nc_prefix(self):
        # It is a switch for user to turn on/off "nc" prefix
        self.device_handler = manager.make_device_handler({'name': 'ericsson', 'with_ns': False})
        session = ncclient.transport.SSHSession(self.device_handler)
        obj = RPC(session, self.device_handler, raise_mode=RaiseMode.ALL, timeout=0)

        expected = """<?xml version="1.0" encoding="UTF-8"?><rpc xmlns="urn:ietf:params:xml:ns:netconf:base:1.0" message-id="%s"><get-config><source><running/></source></get-config></rpc>""" % obj.id

        node = new_ele("get-config")
        child = sub_ele(node, "source")
        sub_ele(child, "running")

        # It is a switch for user to turn on/off "nc" prefix
        rpc_node = obj._wrap(node)
        self.assertEqual(rpc_node, expected)

    def test_rpc_enable_nc_prefix(self):
        # It is a switch for user to turn on/off "nc" prefix
        self.device_handler = manager.make_device_handler({'name': 'ericsson', 'with_ns': True})
        session = ncclient.transport.SSHSession(self.device_handler)
        obj = RPC(session, self.device_handler, raise_mode=RaiseMode.ALL, timeout=0)

        expected = """<?xml version="1.0" encoding="UTF-8"?><nc:rpc xmlns:nc="urn:ietf:params:xml:ns:netconf:base:1.0" message-id="%s"><nc:get-config><nc:source><nc:running/></nc:source></nc:get-config></nc:rpc>""" % obj.id

        node = new_ele("get-config")
        child = sub_ele(node, "source")
        sub_ele(child, "running")

        rpc_node = obj._wrap(node)
        self.assertEqual(rpc_node, expected)

    def test_rpc_enable_nc_prefix_exception(self):
        # invalid value in "with_ns"
        self.device_handler = manager.make_device_handler({'name': 'ericsson', 'with_ns': "Invalid_value"})
        session = ncclient.transport.SSHSession(self.device_handler)
        obj = RPC(session, self.device_handler, raise_mode=RaiseMode.ALL, timeout=0)

        node = new_ele("get-config")
        child = sub_ele(node, "source")
        sub_ele(child, "running")

        self.assertRaises(OperationError, obj._wrap, node)

suite = unittest.TestSuite()
unittest.TextTestRunner().run(suite)




#!/usr/bin/env python

"""
 Listen on TCP port 2200 for incoming SSH session from Junos devices with
 the following ssh outbound configuration and collect host-name and junos-version
 upon connect, then terminate

 lab@router> show configuration system services outbound-ssh
 client outbound-ssh-ncclient {
     device-id vRR;
     services netconf;
     10.0.2.2 port 2200;
  }

 Example:

 $ ./outbound-ssh-ncclient.py
 Listening on port 2200 for incoming sessions ...
 Got a connection from 172.17.0.1:48038!
 MSG DEVICE-CONN-INFO V1 vRR
 Logging in ...
 requesting info...
   Hostname: vRR
    Version: 16.1R3.10
 $
"""


import logging
import socket
import sys
import time

from ncclient import manager
from ncclient.xml_ import *


def listener(port, user, password):
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    s.bind(('', port))
    s.listen(5)
    logging.info('Listening on port %d for incoming sessions ...', port)
    while True:
        client, addr = s.accept()
        logging.info('Got a connection from %s:%d!', addr[0], addr[1])
        launch_junos_proxy(client, addr, user, password)


def launch_junos_proxy(client, addr, user, password):
    val = {
        'MSG-ID': None,
        'MSG-VER': None,
        'DEVICE-ID': None
    }
    msg = ''
    count = 3
    while len(msg) < 100 and count > 0:
        c = client.recv(1)
        c = c.decode()
        if c == '\r':
            continue

        if c == '\n':
            count -= 1
            if msg.find(':'):
                (key, value) = msg.split(': ')
                val[key] = value
                msg = ''
        else:
            msg += c

    logging.info('MSG %s %s %s', val['MSG-ID'], val['MSG-VER'], val['DEVICE-ID'])
    logging.info('Logging in ...')

    sock_fd = client.fileno()
    conn = manager.connect(host=None,
                           sock_fd=sock_fd,
                           username=user,
                           password=password,
                           timeout=60,
                           device_params={'name': 'junos'},
                           hostkey_verify=False)

    rpc = new_ele('get-software-information')

    logging.info('requesting info...')
    result = conn.rpc(rpc)
    logging.info('   Hostname: ' + result.xpath('//software-information/host-name')[0].text)
    logging.info('    Version: ' + result.xpath('//software-information/junos-version')[0].text)
    sys.exit(0)


if __name__ == '__main__':
    LOG_FORMAT = '%(asctime)s %(levelname)s %(filename)s:%(lineno)d %(message)s'
    logging.basicConfig(stream=sys.stdout, level=logging.INFO, format=LOG_FORMAT)

    listener(2200, 'netconf', 'juniper!')



from ncclient.xml_ import BASE_NS_1_0
from .default import DefaultDeviceHandler
from ncclient.operations.third_party.hpcomware.rpc import DisplayCommand, ConfigCommand, Action, Rollback, Save

class HpcomwareDeviceHandler(DefaultDeviceHandler):

    def __init__(self, device_params, ignore_errors=None):
        super(HpcomwareDeviceHandler, self).__init__(device_params, ignore_errors)

    def get_xml_base_namespace_dict(self):
        """
        Base namespace needs a None key.

        See 'nsmap' argument for lxml's Element().

        """
        return {None: BASE_NS_1_0}

    def get_xml_extra_prefix_kwargs(self):
        """
        Return keyword arguments per request, which are applied to Element().

        Mostly, this is a dictionary containing the "nsmap" key.

        See 'nsmap' argument for lxml's Element().

        """
        d = {
                "data": "http://www.hp.com/netconf/data:1.0",
                "config": "http://www.hp.com/netconf/config:1.0",
            }
        d.update(self.get_xml_base_namespace_dict())
        return {"nsmap": d}

    def add_additional_operations(self):
        addtl = {}
        addtl['cli_display'] = DisplayCommand
        addtl['cli_config'] = ConfigCommand
        addtl['action'] = Action
        addtl['rollback'] = Rollback
        addtl['save'] = Save
        return addtl



"""
Handler for Ericsson device specific information.

Note that for proper import, the classname has to be:

    "<Devicename>DeviceHandler"

...where <Devicename> is something like "Default", "Ericsson", etc.

All device-specific handlers derive from the DefaultDeviceHandler, which implements the
generic information needed for interaction with a Netconf server.

"""
from ncclient.xml_ import BASE_NS_1_0
from ncclient.operations.errors import OperationError
from .default import DefaultDeviceHandler


class EricssonDeviceHandler(DefaultDeviceHandler):
    """
    Ericsson handler for device specific information.

    """
    _EXEMPT_ERRORS = []

    def __init__(self, device_params, ignore_errors=None):
        super(EricssonDeviceHandler, self).__init__(device_params, ignore_errors)

    def get_xml_base_namespace_dict(self):
        return {None: BASE_NS_1_0}

    def get_xml_extra_prefix_kwargs(self):
        d = {}
        if self.check_device_params() is False:
            d.update(self.get_xml_base_namespace_dict())
        return {"nsmap": d}

    def check_device_params(self):
        value = self.device_params.get('with_ns')
        if value in [True, False]:
            return value
        elif value is None:
            return False
        else:
            raise OperationError('Invalid "with_ns" value: %s' % value)



# Copyright 2009 Shikhar Bhushan
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ncclient.operations.errors import OperationError
from ncclient.operations.rpc import RPC, RPCReply

from ncclient.xml_ import *
from lxml import etree

from ncclient.operations import util


class WithDefaultsError(OperationError):

    """Invalid 'with-defaults' mode or capability URI"""


class GetReply(RPCReply):

    """Adds attributes for the *data* element to `RPCReply`."""

    def _parsing_hook(self, root):
        self._data = None
        if not self._errors:
            self._data = root.find(qualify("data"))

    @property
    def data_ele(self):
        "*data* element as an :class:`~xml.etree.ElementTree.Element`"
        if not self._parsed:
            self.parse()
        return self._data

    @property
    def data_xml(self):
        "*data* element as an XML string"
        if not self._parsed:
            self.parse()
        return to_xml(self._data)

    data = data_ele
    "Same as :attr:`data_ele`"


class GetSchemaReply(GetReply):
    """Reply for GetSchema called with specific parsing hook."""

    def _parsing_hook(self, root):
        self._data = None
        if not self._errors:
            self._data = root.find(qualify("data", NETCONF_MONITORING_NS)).text


class Get(RPC):

    "The *get* RPC."

    REPLY_CLS = GetReply
    "See :class:`GetReply`."

    def request(self, filter=None, with_defaults=None):
        """Retrieve running configuration and device state information.

        *filter* specifies the portion of the configuration to retrieve (by default entire configuration is retrieved)

        *with_defaults* defines an explicit method of retrieving default values from the configuration (see :rfc:`6243`)

        :seealso: :ref:`filter_params`
        """
        node = new_ele("get")
        if filter is not None:
            node.append(util.build_filter(filter))
        if with_defaults is not None:
            self._assert(":with-defaults")
            _append_with_defaults_mode(
                node,
                with_defaults,
                self._session.server_capabilities,
            )
        return self._request(node)


def _append_with_defaults_mode(node, mode, capabilities):
    _validate_with_defaults_mode(mode, capabilities)
    with_defaults_element = sub_ele_ns(
        node,
        "with-defaults",
        NETCONF_WITH_DEFAULTS_NS,
    )
    with_defaults_element.text = mode


def _validate_with_defaults_mode(mode, capabilities):
    valid_modes = _get_valid_with_defaults_modes(capabilities)
    if mode.strip().lower() not in valid_modes:
        raise WithDefaultsError(
            "Invalid 'with-defaults' mode '{provided}'; the server only "
            "supports the following: {options}".format(
                provided=mode,
                options=', '.join(valid_modes)
            )
        )


def _get_valid_with_defaults_modes(capabilities):
    """Reference: https://tools.ietf.org/html/rfc6243#section-4.3"""
    capability = capabilities[":with-defaults"]

    try:
        valid_modes = [capability.parameters["basic-mode"]]
    except KeyError:
        raise WithDefaultsError(
            "Invalid 'with-defaults' capability URI advertised by the server; "
            "missing 'basic-mode' parameter"
        )

    try:
        also_supported = capability.parameters["also-supported"]
    except KeyError:
        return valid_modes

    valid_modes.extend(also_supported.split(","))

    return valid_modes


class GetConfig(RPC):

    """The *get-config* RPC."""

    REPLY_CLS = GetReply
    """See :class:`GetReply`."""

    def request(self, source, filter=None, with_defaults=None):
        """Retrieve all or part of a specified configuration.

        *source* name of the configuration datastore being queried

        *filter* specifies the portion of the configuration to retrieve (by default entire configuration is retrieved)

        *with_defaults* defines an explicit method of retrieving default values from the configuration (see :rfc:`6243`)

        :seealso: :ref:`filter_params`"""
        node = new_ele("get-config")
        node.append(util.datastore_or_url("source", source, self._assert))
        if filter is not None:
            node.append(util.build_filter(filter))
        if with_defaults is not None:
            self._assert(":with-defaults")
            _append_with_defaults_mode(
                node,
                with_defaults,
                self._session.server_capabilities,
            )
        return self._request(node)

class GetSchema(RPC):

    """The *get-schema* RPC."""

    REPLY_CLS = GetSchemaReply
    """See :class:`GetReply`."""

    def request(self, identifier, version=None, format=None):
        """Retrieve a named schema, with optional revision and type.

        *identifier* name of the schema to be retrieved

        *version* version of schema to get

        *format* format of the schema to be retrieved, yang is the default

        :seealso: :ref:`filter_params`"""
        self._huge_tree = True
        node = etree.Element(qualify("get-schema",NETCONF_MONITORING_NS))
        if identifier is not None:
            elem = etree.Element(qualify("identifier",NETCONF_MONITORING_NS))
            elem.text = identifier
            node.append(elem)
        if version is not None:
            elem = etree.Element(qualify("version",NETCONF_MONITORING_NS))
            elem.text = version
            node.append(elem)
        if format is not None:
            elem = etree.Element(qualify("format",NETCONF_MONITORING_NS))
            elem.text = format
            node.append(elem)
        return self._request(node)

class Dispatch(RPC):

    """Generic retrieving wrapper"""

    REPLY_CLS = RPCReply
    """See :class:`RPCReply`."""

    def request(self, rpc_command, source=None, filter=None):
        """
        *rpc_command* specifies rpc command to be dispatched either in plain text or in xml element format (depending on command)

        *source* name of the configuration datastore being queried

        *filter* specifies the portion of the configuration to retrieve (by default entire configuration is retrieved)

        :seealso: :ref:`filter_params`

        Examples of usage::

            dispatch('clear-arp-table')

        or dispatch element like ::

            xsd_fetch = new_ele('get-xnm-information')
            sub_ele(xsd_fetch, 'type').text="xml-schema"
            sub_ele(xsd_fetch, 'namespace').text="junos-configuration"
            dispatch(xsd_fetch)
        """


        if etree.iselement(rpc_command):
            node = rpc_command
        else:
            node = new_ele(rpc_command)
        if source is not None:
            node.append(util.datastore_or_url("source", source, self._assert))
        if filter is not None:
            node.append(util.build_filter(filter))

        return self._request(node)



import unittest
from ncclient.devices.iosxe import *
from ncclient.xml_ import new_ele
from ncclient.xml_ import qualify
from ncclient.xml_ import validated_element


CFG_BROKEN = """
<config>
  <native xmlns="http://cisco.com/ns/yang/Cisco-IOS-XE-native">
    <hostname>tl-einarnn-c8kv</hostname>
  </native>
</config>
"""


class TestIosxeDevice(unittest.TestCase):
    
    def setUp(self):
        self.obj = IosxeDeviceHandler({'name': 'iosxe'})

    def test_add_additional_operations(self):
        expected = {'save_config': SaveConfig}
        self.assertDictEqual(expected, self.obj.add_additional_operations())

    def test_add_additional_ssh_connect_params(self):
        expected = dict()
        expected["unknown_host_cb"] = iosxe_unknown_host_cb
        actual = dict()
        self.obj.add_additional_ssh_connect_params(actual)
        self.assertDictEqual(expected, actual)

    def test_csr_unknown_host_cb(self):
        self.assertTrue(iosxe_unknown_host_cb('host', 'fingerprint'))

    def test_iosxe_transform_edit_config(self):
        node = new_ele("edit-config")
        node.append(validated_element(CFG_BROKEN, ("config", qualify("config"))))
        node = self.obj.transform_edit_config(node)
        config_nodes = node.findall('./config')
        self.assertTrue(len(config_nodes) == 0)



import unittest
from ncclient.devices.huaweiyang import *


class TestHuaweiyangDevice(unittest.TestCase):

    def setUp(self):
        self.obj = HuaweiyangDeviceHandler({'name': 'huaweiyang'})

    def test_get_capabilities(self):
        expected = ['urn:ietf:params:netconf:base:1.0', 'urn:ietf:params:netconf:base:1.1']
        self.assertListEqual(expected, self.obj.get_capabilities())

    def test_get_xml_base_namespace_dict(self):
        expected = {None: BASE_NS_1_0}
        self.assertDictEqual(expected, self.obj.get_xml_base_namespace_dict())

    def test_get_xml_extra_prefix_kwargs(self):
        expected = dict()
        expected['nsmap'] = self.obj.get_xml_base_namespace_dict()
        self.assertDictEqual(expected, self.obj.get_xml_extra_prefix_kwargs())



#!/usr/bin/env python
import logging
import os

from ncclient import manager
from ncclient.xml_ import *


def connect(host, port, user, password):
    conn = manager.connect(host=host,
                           port=port,
                           username=user,
                           password=password,
                           timeout=1800,
                           device_params={'name': 'junos'},
                           hostkey_verify=False)

    # https://www.juniper.net/documentation/en_US/junos/topics/task/operational/junos-xml-protocol-requesting-xml-schema.html
    logging.info("Requesting an XML Schema for the Configuration Hierarchy")
    logging.info("This may take several (even 10+) minutes")

    rpc = """
    <get-xnm-information>
        <type>xml-schema</type>
        <namespace>junos-configuration</namespace>
    </get-xnm-information>"""

    result = conn.rpc(rpc)
    with open('schema.txt', 'w') as fh:
        # Note: using NCElement's __str__() for python version independent conversion to string
        fh.write(str(result))
        logging.info('schema.txt is written to directory %s', os.getcwd())


if __name__ == '__main__':
    LOG_FORMAT = '%(asctime)s %(levelname)s %(filename)s:%(lineno)d %(message)s'
    logging.basicConfig(stream=sys.stdout, level=logging.INFO, format=LOG_FORMAT)

    connect('router', 830, 'netconf', 'juniper!')



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import socket
import threading
from io import BytesIO as StringIO

from socket import AF_UNIX, SOCK_STREAM

from ncclient.capabilities import Capabilities
from ncclient.logging_ import SessionLoggerAdapter
from ncclient.transport.errors import UnixSocketError
from ncclient.transport.session import Session
from ncclient.transport.parser import DefaultXMLParser

logger = logging.getLogger("ncclient.transport.unix")

DEFAULT_TIMEOUT = 120

BUF_SIZE = 4096

class UnixSocketSession(Session):

    "Implements a NETCONF Session over Unix Socket on local machine."

    def __init__(self, device_handler):
        capabilities = Capabilities(device_handler.get_capabilities())
        Session.__init__(self, capabilities)
        self._connected = False
        self._socket = None
        self._buffer = StringIO()
        self._device_handler = device_handler
        self._message_list = []
        self._closing = threading.Event()
        self.parser = DefaultXMLParser(self)
        self.logger = SessionLoggerAdapter(logger, {'session': self})

    def _dispatch_message(self, raw):
        self.logger.info("Received message from host")
        self.logger.debug("Received:\n%s", raw)
        return super(UnixSocketSession, self)._dispatch_message(raw)

    def close(self):
        self._closing.set()
        self._socket.close()
        self._connected = False

    def connect(self, path=None, timeout=DEFAULT_TIMEOUT):
        sock = socket.socket(AF_UNIX, SOCK_STREAM)
        sock.settimeout(timeout)

        try:
            sock.connect(path)
        except Exception:
            raise UnixSocketError("Could not connect to %s" % path)

        self._socket = sock
        self._connected = True
        self._post_connect()

    def _transport_read(self):
        return self._socket.recv(BUF_SIZE)

    def _transport_write(self, data):
        return self._socket.send(data)

    def _transport_register(self, selector, event):
        selector.register(self._socket, event)

    def _send_ready(self):
        # In contrast to Paramiko's `Channel`, pure python sockets do not
        # expose `send_ready()` function.
        return True



import unittest
from ncclient import devices

class  TestGetSupportedDevices(unittest.TestCase):

    def test_get_supported_devices(self):
        supported_devices = devices.get_supported_devices()
        self.assertEqual(sorted(supported_devices), sorted(('junos',
                                                            'csr',
                                                            'ciena',
                                                            'nexus',
                                                            'iosxr',
                                                            'iosxe',
                                                            'huawei',
                                                            'huaweiyang',
                                                            'alu',
                                                            'h3c',
                                                            'hpcomware',
                                                            'sros',
                                                            'default')))

    def test_get_supported_device_labels(self):
        supported_device_labels = devices.get_supported_device_labels()
        self.assertEqual(supported_device_labels, {'junos':'Juniper',
                                                   'csr':'Cisco CSR1000v',
                                                   'ciena': 'Ciena',
                                                   'nexus':'Cisco Nexus',
                                                   'iosxr':'Cisco IOS XR',
                                                   'iosxe':'Cisco IOS XE',
                                                   'huawei':'Huawei',
                                                   'huaweiyang':'Huawei',
                                                   'alu':'Alcatel Lucent',
                                                   'h3c':'H3C',
                                                   'hpcomware':'HP Comware',
                                                   'sros':'Nokia SR OS',
                                                   'default':'Server or anything not in above'})




import unittest
try:
    from unittest.mock import patch  # Python 3.4 and later
except ImportError:
    from mock import patch
from ncclient.transport.session import *
from ncclient.devices.junos import JunosDeviceHandler
try:
    from Queue import Queue, Empty
except ImportError:
    from queue import Queue, Empty
import logging



rpc_reply = """<rpc-reply xmlns:junos="http://xml.juniper.net/junos/12.1X46/junos" attrib1 = "test">
    <software-information>
        <host-name>R1</host-name>
        <product-model>firefly-perimeter</product-model>
        <product-name>firefly-perimeter</product-name>
        <package-information>
            <name>junos</name>
            <comment>JUNOS Software Release [12.1X46-D10.2]</comment>
        </package-information>
    </software-information>
    <cli>
        <banner></banner>
    </cli>
</rpc-reply>
"""

hello_rpc_reply = """<hello>
<capabilities>
<capability>candidate</capability>
<capability>validate</capability>
</capabilities>
<session-id>s001</session-id>
</hello>
"""

notification="""
   <notification
      xmlns="urn:ietf:params:xml:ns:netconf:notification:1.0">
      <eventTime>2007-07-08T00:01:00Z</eventTime>
      <event xmlns="http://example.com/event/1.0">
         <eventClass>fault</eventClass>
         <reportingEntity>
             <card>Ethernet0</card>
         </reportingEntity>
         <severity>major</severity>
       </event>
   </notification>
"""

class TestSession(unittest.TestCase):

    @patch('ncclient.transport.session.HelloHandler.callback')
    def test_dispatch_message(self, mock_handler):
        cap = [':candidate']
        obj = Session(cap)
        device_handler = JunosDeviceHandler({'name': 'junos'})
        obj._device_handler = device_handler
        listener = HelloHandler(None, None)
        obj._listeners.add(listener)
        obj._dispatch_message(rpc_reply)
        mock_handler.assert_called_once_with(parse_root(rpc_reply), rpc_reply)

    @patch('ncclient.transport.session.parse_root')
    @patch('ncclient.logging_.SessionLoggerAdapter.error')
    def test_dispatch_message_error(self, mock_log, mock_parse_root):
        mock_parse_root.side_effect = Exception
        cap = [':candidate']
        obj = Session(cap)
        device_handler = JunosDeviceHandler({'name': 'junos'})
        obj._device_handler = device_handler
        listener = HelloHandler(None, None)
        obj._listeners.add(listener)
        obj._dispatch_message(rpc_reply)
        self.assertNotEqual(
            mock_log.call_args_list[0][0][0].find("error parsing dispatch message"), -1)

    @patch('ncclient.transport.session.parse_root')
    @patch('ncclient.devices.junos.JunosDeviceHandler.handle_raw_dispatch')
    @patch('ncclient.transport.session.HelloHandler.errback')
    @patch('ncclient.logging_.SessionLoggerAdapter.debug')
    def test_dispatch_message_error2(self, mock_log, mock_errback,
                                     mock_handle_raw_dispatch, mock_parse_root):
        mock_parse_root.side_effect = Exception
        mock_handle_raw_dispatch.return_value = Exception()
        mock_errback.side_effect = Exception
        cap = [':candidate']
        obj = Session(cap)
        device_handler = JunosDeviceHandler({'name': 'junos'})
        obj._device_handler = device_handler
        listener = HelloHandler(None, None)
        obj._listeners.add(listener)
        obj._dispatch_message(rpc_reply)
        mock_handle_raw_dispatch.assert_called_once_with(rpc_reply)
        self.assertEqual(
            mock_log.call_args_list[0][0][0].find("error dispatching to"), -1)

    @patch('ncclient.transport.session.HelloHandler.errback')
    def test_dispatch_error(self, mock_handler):
        cap = [':candidate']
        obj = Session(cap)
        listener = HelloHandler(None, None)
        obj._listeners.add(listener)
        obj._dispatch_error("Error")
        mock_handler.assert_called_once_with("Error")

    @patch('ncclient.logging_.SessionLoggerAdapter.info')
    @patch('ncclient.transport.session.Thread.start')
    @patch('ncclient.transport.session.Event')
    def test_post_connect(self, mock_lock, mock_handler, mock_log):
        cap = [':candidate']
        obj = Session(cap)
        device_handler = JunosDeviceHandler({'name': 'junos'})
        obj._device_handler = device_handler
        obj._connected = True
        obj._id = 100
        obj._server_capabilities = cap
        obj._post_connect()
        log_args = mock_log.call_args_list[0][0]
        self.assertNotEqual(log_args[0].find("initialized"), -1)
        self.assertNotEqual(log_args[0].find("session-id="), -1)
        self.assertEqual(log_args[1], 100)
        self.assertNotEqual(
            log_args[0].find("server_capabilities="), -1)
        self.assertEqual(log_args[2], [':candidate'])

    @patch('ncclient.logging_.SessionLoggerAdapter.info')
    @patch('ncclient.transport.session.Thread.start')
    @patch('ncclient.transport.session.Event')
    def test_post_connect2(self, mock_lock, mock_handler, mock_log):
        cap = ['urn:ietf:params:netconf:base:1.1']
        obj = Session(cap)
        device_handler = JunosDeviceHandler({'name': 'junos'})
        obj._device_handler = device_handler
        obj._connected = True
        obj._id = 100
        obj._server_capabilities = cap
        obj._post_connect()
        log_args = mock_log.call_args_list[0][0]
        self.assertNotEqual(log_args[0].find("initialized"), -1)
        self.assertNotEqual(log_args[0].find("session-id="), -1)
        self.assertEqual(log_args[1], 100)
        self.assertNotEqual(
            log_args[0].find("server_capabilities="), -1)
        self.assertEqual(log_args[2], ['urn:ietf:params:netconf:base:1.1'])

    def test_add_listener(self):
        cap = [':candidate']
        obj = Session(cap)
        listener = HelloHandler(None, None)
        obj.add_listener(listener)
        self.assertTrue(listener in obj._listeners)

    def test_add_listener_exception(self):
        cap = [':candidate']
        obj = Session(cap)
        listener = Session(None)
        self.assertRaises(SessionError,
            obj.add_listener, listener)

    def test_remove_listener(self):
        cap = [':candidate']
        obj = Session(cap)
        listener = HelloHandler(None, None)
        obj._listeners.add(listener)
        obj.remove_listener(listener)
        self.assertEqual(len(obj._listeners), 0)

    def test_get_listener_instance(self):
        cap = [':candidate']
        obj = Session(cap)
        listener = HelloHandler(None, None)
        obj._listeners.add(listener)
        ret = obj.get_listener_instance(HelloHandler)
        self.assertEqual(ret, listener)

    def test_send_connected(self):
        cap = [':candidate']
        obj = Session(cap)
        obj._connected = True
        obj.send("Hello World")
        self.assertEqual("Hello World", obj._q.get())

    def test_send_disconnected(self):
        cap = [':candidate']
        obj = Session(cap)
        obj._connected = False
        self.assertRaises(TransportError,
            obj.send, "Hello World")

    def test_connected(self):
        cap = [':candidate']
        obj = Session(cap)
        obj._connected = True
        self.assertTrue(obj.connected)

    def test_client_capability(self):
        cap = [':validate']
        obj = Session(cap)
        self.assertEqual(obj.client_capabilities, cap)

    def test_server_capability(self):
        cap = [':validate']
        obj = Session(cap)
        obj._server_capabilities = cap
        self.assertEqual(obj.server_capabilities, cap)

    def test_id(self):
        cap = [':validate']
        obj = Session(cap)
        obj._id = "1000"
        self.assertEqual(obj.id, "1000")

    def test_hello_handler_callback_ok(self):

        def ok_cb(id, capabilities):
            self._id = id
            self._server_capabilities = capabilities

        def err_cb(err):
            self.error = err

        listener = HelloHandler(ok_cb, err_cb)
        listener.callback(parse_root(hello_rpc_reply), hello_rpc_reply)
        caps = ["candidate", "validate"]
        self.assertEqual(self._id, "s001")
        self.assertEqual(
            self._server_capabilities._dict,
            Capabilities(caps)._dict)

    @patch('ncclient.transport.session.HelloHandler.parse')
    def test_hello_handler_callback_error(self, mock_parse):

        def ok_cb(id, capabilities):
            self._id = id
            self._server_capabilities = capabilities

        def err_cb(err):
            self.error = "Error"

        mock_parse.side_effect = Exception
        listener = HelloHandler(ok_cb, err_cb)
        listener.callback(parse_root(hello_rpc_reply), hello_rpc_reply)
        self.assertEqual(self.error, "Error")

    def test_hello_handler_build(self):
        cap = [':candidate', ':validate']
        listener = HelloHandler(None, None)
        result = listener.build(cap, None)
        node = new_ele("hello")
        caps = sub_ele(node, "capabilities")
        sub_ele(caps, "capability").text = ":candidate"
        sub_ele(caps, "capability").text = ":validate"
        node = to_xml(node)
        self.assertEqual(node, result)

    def test_hello_handler_parse(self):
        cap = [':candidate']
        obj = Session(cap)
        listener = HelloHandler(None, None)
        obj._listeners.add(listener)
        id, capabilities = listener.parse(hello_rpc_reply)
        self.assertEqual(id, "s001")
        caps = ["candidate", "validate"]
        self.assertEqual(capabilities._dict, Capabilities(caps)._dict)

    def test_notification_handler_valid_notification(self):
        q = Queue()
        listener = NotificationHandler(q)
        listener.callback(parse_root(notification), notification)
        notif = q.get_nowait()
        self.assertEqual(notif.notification_xml, notification)
        self.assertRaises(Empty, q.get_nowait)

    def test_notification_handler_non_notification(self):
        q = Queue()
        listener = NotificationHandler(q)
        # This handler should ignore things that aren't notifications
        listener.callback(parse_root(rpc_reply), rpc_reply)
        self.assertRaises(Empty, q.get_nowait)

    def test_take_notification(self):
        cap = [':candidate']
        obj = Session(cap)
        obj._notification_q.put('Test object')
        self.assertEqual(obj.take_notification(block=False, timeout=None),
                         'Test object')
        self.assertEqual(obj.take_notification(block=False, timeout=None),
                         None)



# Copyright 2018 Nitin Kumar

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from threading import Lock

import difflib

from xml.sax.handler import ContentHandler
from lxml import etree
from lxml.builder import E
from xml.sax._exceptions import SAXParseException
from xml.sax import make_parser

from ncclient.transport.parser import DefaultXMLParser
from ncclient.operations import rpc
from ncclient.transport.parser import SAXFilterXMLNotFoundError
from ncclient.transport.parser import MSG_DELIM, MSG_DELIM_LEN
from ncclient.operations.errors import OperationError

import logging
logger = logging.getLogger("ncclient.transport.third_party.junos.parser")

from ncclient.xml_ import BASE_NS_1_0

RPC_REPLY_END_TAG = "</rpc-reply>"
RFC_RPC_REPLY_END_TAG = "</nc:rpc-reply>"


class JunosXMLParser(DefaultXMLParser):
    def __init__(self, session):
        super(JunosXMLParser, self).__init__(session)
        self._session = session
        self.sax_parser = make_parser()
        self.sax_parser.setContentHandler(SAXParser(session))

    def parse(self, data):
        try:
            self.sax_parser.feed(data)
        except SAXParseException:
            self._delimiter_check(data)
        except SAXFilterXMLNotFoundError:
            self.logger.debug('Missing SAX filter_xml. Switching from sax to dom parsing')
            self._session.parser = DefaultXMLParser(self._session)
            if not isinstance(data, bytes):
                data = str.encode(data)
            self._session._buffer.write(data)
        finally:
            self._parse10()

    def _delimiter_check(self, data):
        """
        SAX parser throws SAXParseException exception, if there is extra data
        after MSG_DELIM

        :param data: content read by select loop
        :return: None
        """
        data = data.decode('UTF-8')
        if MSG_DELIM in data:
            # need to feed extra data after MSG_DELIM
            msg, delim, remaining = data.partition(MSG_DELIM)
            self._session._buffer.seek(0, os.SEEK_END)
            self._session._buffer.write(delim.encode())
            if remaining.strip() != '':
                self._session._buffer.write(remaining.encode())
            # we need to renew parser, as old parser is gone.
            self.sax_parser = make_parser()
            self.sax_parser.setContentHandler(SAXParser(self._session))
        elif RPC_REPLY_END_TAG in data or RFC_RPC_REPLY_END_TAG in data:
            tag = RPC_REPLY_END_TAG if RPC_REPLY_END_TAG in data else \
                RFC_RPC_REPLY_END_TAG
            logger.warning("Check for rpc reply end tag within data received: %s" % data)
            msg, delim, remaining = data.partition(tag)
            self._session._buffer.seek(0, os.SEEK_END)
            self._session._buffer.write(remaining.encode())
        else:
            logger.warning("Check if end delimiter is split within data received: %s" % data)
            # When data is "-reply/>]]>" or "]]>"
            # Data is not full MSG_DELIM, So check if last rpc reply is complete.
            # if then, wait for next iteration of data and do a recursive call to
            # _delimiter_check for MSG_DELIM check
            buf = self._session._buffer
            buf.seek(buf.tell() - len(RFC_RPC_REPLY_END_TAG) - MSG_DELIM_LEN)
            rpc_response_last_msg = buf.read().decode('UTF-8').replace('\n', '')
            if RPC_REPLY_END_TAG in rpc_response_last_msg or \
                    RFC_RPC_REPLY_END_TAG in rpc_response_last_msg:
                tag = RPC_REPLY_END_TAG if RPC_REPLY_END_TAG in \
                                           rpc_response_last_msg else \
                    RFC_RPC_REPLY_END_TAG
                # rpc_response_last_msg and data can be overlapping
                match_obj = difflib.SequenceMatcher(None, rpc_response_last_msg,
                                                    data).get_matching_blocks()
                if match_obj:
                    # 0 means second string match start from beginning, hence
                    # there is a overlap
                    if match_obj[0].b == 0:
                        # matching char are of match_obj[0].size
                        self._delimiter_check((rpc_response_last_msg +
                                               data[match_obj[0].size:]).encode())
                    else:
                        data = rpc_response_last_msg+data
                        if MSG_DELIM in data:
                            # there can be residual end delimiter chars in buffer.
                            # as first if condition will add full delimiter, so clean
                            # it off
                            clean_up = len(rpc_response_last_msg) - (
                                    rpc_response_last_msg.find(tag) + len(tag))
                            self._session._buffer.truncate(buf.tell() - clean_up)
                            self._delimiter_check(data.encode())
                        else:
                            self._delimiter_check((rpc_response_last_msg + data).encode())


def __dict_replace(s, d):
    """Replace substrings of a string using a dictionary."""
    for key, value in d.items():
        s = s.replace(key, value)
    return s


def _get_sax_parser_root(xml):
    """
    This function does some validation and rule check of xmlstring
    :param xml: string or object to be used in parsing reply
    :return: lxml object
    """
    if isinstance(xml, etree._Element):
        root = xml
    else:
        root = etree.fromstring(xml)
    return root


def escape(data, entities={}):
    """Escape &, <, and > in a string of data.

    You can escape other strings of data by passing a dictionary as
    the optional entities parameter.  The keys and values must all be
    strings; each key will be replaced with its corresponding value.
    """

    # must do ampersand first
    data = data.replace("&", "&amp;")
    data = data.replace(">", "&gt;")
    data = data.replace("<", "&lt;")
    if entities:
        data = __dict_replace(data, entities)
    return data


def quoteattr(data, entities={}):
    """Escape and quote an attribute value.

    Escape &, <, and > in a string of data, then quote it for use as
    an attribute value.  The \" character will be escaped as well, if
    necessary.

    You can escape other strings of data by passing a dictionary as
    the optional entities parameter.  The keys and values must all be
    strings; each key will be replaced with its corresponding value.
    """
    # entities = entities.copy()
    entities.update({'\n': '&#10;', '\r': '&#13;', '\t':'&#9;'})
    data = escape(data, entities)
    if '"' in data:
        if "'" in data:
            data = '"%s"' % data.replace('"', "&quot;")
        else:
            data = "'%s'" % data
    else:
        data = '"%s"' % data
    return data


class SAXParser(ContentHandler):
    def __init__(self, session):
        ContentHandler.__init__(self)
        self._currenttag = None
        self._ignoretag = None
        self._defaulttags = []
        self._session = session
        self._validate_reply_and_sax_tag = False
        self._lock = Lock()
        self.nc_namespace = None

    def startElement(self, tag, attributes):
        if tag in ['rpc-reply', 'nc:rpc-reply']:
            if tag == 'nc:rpc-reply':
                self.nc_namespace = BASE_NS_1_0
            # in case last rpc called used sax parsing and error'd out
            # without resetting use_filer in endElement rpc-reply check
            with self._lock:
                listeners = list(self._session._listeners)
            rpc_reply_listener = [i for i in listeners if
                                 isinstance(i, rpc.RPCReplyListener)]
            rpc_msg_id = attributes._attrs['message-id']
            if rpc_msg_id in rpc_reply_listener[0]._id2rpc:
                rpc_reply_handler = rpc_reply_listener[0]._id2rpc[rpc_msg_id]
                if hasattr(rpc_reply_handler, '_filter_xml') and \
                        rpc_reply_handler._filter_xml is not None:
                    self._cur = self._root = _get_sax_parser_root(
                        rpc_reply_handler._filter_xml)
                else:
                    raise SAXFilterXMLNotFoundError(rpc_reply_handler)
            else:
                raise OperationError("Unknown 'message-id': %s" % rpc_msg_id)
        if self._ignoretag is not None:
            return

        if self._cur == self._root and self._cur.tag == tag:
            node = self._root
        else:
            node = self._cur.find(tag, namespaces={"nc": self.nc_namespace})

        if self._validate_reply_and_sax_tag:
            if tag != self._root.tag:
                self._write_buffer(tag, format_str='<{}>\n')
                self._cur = E(tag, self._cur)
            else:
                self._write_buffer(tag, format_str='<{}{}>', **attributes)
                self._cur = node
                self._currenttag = tag
            self._validate_reply_and_sax_tag = False
            self._defaulttags.append(tag)
        elif node is not None:
            self._write_buffer(tag, format_str='<{}{}>', **attributes)
            self._cur = node
            self._currenttag = tag
        elif tag in ['rpc-reply', 'nc:rpc-reply']:
            self._write_buffer(tag, format_str='<{}{}>', **attributes)
            self._defaulttags.append(tag)
            self._validate_reply_and_sax_tag = True
        else:
            self._currenttag = None
            self._ignoretag = tag

    def endElement(self, tag):
        if self._ignoretag == tag:
            self._ignoretag = None

        if tag in self._defaulttags:
            self._write_buffer(tag, format_str='</{}>\n')
        elif self._cur.tag == tag:
            self._write_buffer(tag, format_str='</{}>\n')
            self._cur = self._cur.getparent()

        self._currenttag = None

    def characters(self, content):
        if self._currenttag is not None:
            self._write_buffer(content, format_str='{}')

    def _write_buffer(self, content, format_str, **kwargs):
        # print(content, format_str, kwargs)
        self._session._buffer.seek(0, os.SEEK_END)
        attrs = ''
        for (name, value) in kwargs.items():
            attr = ' {}={}'.format(name, quoteattr(value))
            attrs = attrs + attr
        data = format_str.format(escape(content), attrs)
        self._session._buffer.write(str.encode(data))



#!/usr/bin/env python

import json
import logging

from ncclient import manager
from ncclient.xml_ import *


def connect(host, port, user, password):
    conn = manager.connect(host=host,
                           port=port,
                           username=user,
                           password=password,
                           timeout=60,
                           device_params={'name': 'junos'},
                           hostkey_verify=False)

    result_xml = conn.get_configuration(format='xml')
    logging.info(result_xml)

    result_json = conn.get_configuration(format='json')
    payload = json.loads(result_json.xpath('.')[0].text)
    logging.info(payload)
    logging.info(payload['configuration'][0]['system'][0]['services'])

    result_text = conn.get_configuration(format='text')
    logging.info(result_text.xpath('configuration-text')[0].text)

    logging.info('Version')
    logging.info('*' * 30)
    logging.info(result_xml.xpath('configuration/version')[0].text)

    config_filter = new_ele('configuration')
    system_ele = sub_ele(config_filter, 'system')
    sub_ele(system_ele, 'login')

    result_filter = conn.get_configuration(format='xml', filter=config_filter)
    logging.info(result_filter)

    logging.info('Configured Interfaces...')
    interfaces = result_xml.xpath('configuration/interfaces/interface')
    for i in interfaces:
        interface = i.xpath('name')[0].text
        ip = []
        for name in i.xpath('unit/family/inet/address/name'):
            ip.append(name.text)
        logging.info(' %s %s', interface, ip)


if __name__ == '__main__':
    LOG_FORMAT = '%(asctime)s %(levelname)s %(filename)s:%(lineno)d %(message)s'
    logging.basicConfig(stream=sys.stdout, level=logging.INFO, format=LOG_FORMAT)

    connect('router', 830, 'netconf', 'juniper!')



import unittest
try:
    from unittest.mock import patch  # Python 3.4 and later
except ImportError:
    from mock import patch
from ncclient import manager
import ncclient.transport
from ncclient.operations.third_party.nexus.rpc import *


class TestRPC(unittest.TestCase):
    def setUp(self):
        self.device_handler = manager.make_device_handler({'name': 'nexus'})
    
    @patch('ncclient.operations.third_party.nexus.rpc.RPC._request')
    def test_execCommand(self, mock_request):
        mock_request.return_value = 'nexus'
        expected = 'nexus'
        session = ncclient.transport.SSHSession(self.device_handler)
        obj = ExecCommand(session, self.device_handler)
        cmds = 'show devices-name'
        actual = obj.request(cmds=cmds)
        self.assertEqual(expected, actual)

        commands = [cmd for cmd in cmds]
        actual = obj.request(cmds=commands)
        self.assertEqual(expected, actual)



from ncclient.operations.third_party.juniper.rpc import *
import json
import unittest
try:
    from unittest.mock import patch  # Python 3.4 and later
except ImportError:
    from mock import patch
from ncclient import manager
import ncclient.manager
import ncclient.transport
from ncclient.xml_ import *
from ncclient.operations import RaiseMode
from xml.etree import ElementTree


class TestRPC(unittest.TestCase):

    @patch('ncclient.transport.SSHSession')
    @patch('ncclient.operations.third_party.juniper.rpc.RPC._request')
    def test_command(self, mock_request, mock_session):
        device_handler = manager.make_device_handler({'name': 'junos'})
        session = ncclient.transport.SSHSession(device_handler)
        obj = Command(session, device_handler, raise_mode=RaiseMode.ALL)
        command = 'show system users'
        format = 'text'
        obj.request(command=command, format=format)
        node = new_ele('command', {'format': format})
        node.text = command
        call = mock_request.call_args_list[0][0][0]
        self.assertEqual(call.tag, node.tag)
        self.assertEqual(call.text, node.text)

    @patch('ncclient.transport.SSHSession')
    @patch('ncclient.operations.third_party.juniper.rpc.RPC._request')
    def test_getconf(self, mock_request, mock_session):
        device_handler = manager.make_device_handler({'name': 'junos'})
        session = ncclient.transport.SSHSession(device_handler)
        obj = GetConfiguration(
            session,
            device_handler,
            raise_mode=RaiseMode.ALL)
        root_filter = new_ele('filter')
        config_filter = sub_ele(root_filter, 'configuration')
        system_filter = sub_ele(config_filter, 'system')
        obj.request(format='xml', filter=system_filter)
        node = new_ele('get-configuration', {'format': 'xml'})
        node.append(system_filter)
        call = mock_request.call_args_list[0][0][0]
        self.assertEqual(call.tag, node.tag)
        self.assertEqual(call.attrib, node.attrib)

    @patch('ncclient.transport.SSHSession')
    @patch('ncclient.operations.third_party.juniper.rpc.RPC._request')
    def test_getconf_text_sets_huge_tree(self, mock_request, mock_session):
        device_handler = manager.make_device_handler({'name': 'junos'})
        session = ncclient.transport.SSHSession(device_handler)
        obj = GetConfiguration(
            session,
            device_handler,
            raise_mode=RaiseMode.ALL)
        obj.huge_tree = False
        obj.request(format='text')
        self.assertTrue(obj.huge_tree)

    @patch('ncclient.transport.SSHSession')
    @patch('ncclient.operations.third_party.juniper.rpc.RPC._request')
    def test_loadconf_xml(self, mock_request, mock_session):
        device_handler = manager.make_device_handler({'name': 'junos'})
        session = ncclient.transport.SSHSession(device_handler)
        obj = LoadConfiguration(
            session,
            device_handler,
            raise_mode=RaiseMode.ALL)
        root_config = new_ele('configuration')
        system_config = sub_ele(root_config, 'system')
        location_config = sub_ele(system_config, 'location')
        floor_config = sub_ele(location_config, 'floor').text = "7"
        obj.request(format='xml', config=root_config)
        node = new_ele('load-configuration', {'format': 'xml', 'action': 'merge'})
        node.append(root_config)
        call = mock_request.call_args_list[0][0][0]
        self.assertEqual(call.tag, node.tag)
        self.assertEqual(call.attrib, node.attrib)

    @patch('ncclient.transport.SSHSession')
    @patch('ncclient.operations.third_party.juniper.rpc.RPC._request')
    def test_loadconf_json(self, mock_request, mock_session):
        device_handler = manager.make_device_handler({'name': 'junos'})
        session = ncclient.transport.SSHSession(device_handler)
        obj = LoadConfiguration(
            session,
            device_handler,
            raise_mode=RaiseMode.ALL)
        location = '{ "configuration": { "system": { "location": { "floor": "7" }}}}'
        config_json = json.loads(location)
        config = json.dumps(config_json)
        obj.request(format='json', action='merge', config=config)
        node = new_ele('load-configuration', {'format': 'json', 'action': 'merge'})
        sub_ele(node, 'configuration-json').text = config
        call = mock_request.call_args_list[0][0][0]
        self.assertEqual(call.tag, node.tag)
        self.assertEqual(call.attrib, node.attrib)

    @patch('ncclient.transport.SSHSession')
    @patch('ncclient.operations.third_party.juniper.rpc.RPC._request')
    def test_loadconf_set(self, mock_request, mock_session):
        device_handler = manager.make_device_handler({'name': 'junos'})
        session = ncclient.transport.SSHSession(device_handler)
        obj = LoadConfiguration(
            session,
            device_handler,
            raise_mode=RaiseMode.ALL)
        config = 'set system location floor 7'
        obj.request(format='text', action='set', config=config)
        node = new_ele('load-configuration', {'format': 'text', 'action': 'set'})
        sub_ele(node, 'configuration-set').text = config
        call = mock_request.call_args_list[0][0][0]
        self.assertEqual(call.tag, node.tag)
        self.assertEqual(call.attrib, node.attrib)

    @patch('ncclient.transport.SSHSession')
    @patch('ncclient.operations.third_party.juniper.rpc.RPC._request')
    def test_loadconf_text(self, mock_request, mock_session):
        device_handler = manager.make_device_handler({'name': 'junos'})
        session = ncclient.transport.SSHSession(device_handler)
        obj = LoadConfiguration(
            session,
            device_handler,
            raise_mode=RaiseMode.ALL)
        config = 'system { location floor 7; }'
        obj.request(format='text', action='merge', config=config)
        node = new_ele('load-configuration', {'format': 'text', 'action': 'merge'})
        sub_ele(node, 'configuration-text').text = config
        call = mock_request.call_args_list[0][0][0]
        self.assertEqual(call.tag, node.tag)
        self.assertEqual(call.attrib, node.attrib)

    @patch('ncclient.transport.SSHSession')
    @patch('ncclient.operations.third_party.juniper.rpc.RPC._request')
    def test_loadconf_list(self, mock_request, mock_session):
        device_handler = manager.make_device_handler({'name': 'junos'})
        session = ncclient.transport.SSHSession(device_handler)
        obj = LoadConfiguration(
            session,
            device_handler,
            raise_mode=RaiseMode.ALL)
        config = ['set system location floor 7', 'set system location rack 3']
        obj.request(format='text', action='set', config=config)
        node = new_ele('load-configuration', {'format': 'text', 'action': 'set'})
        sub_ele(node, 'configuration-set').text = '\n'.join(config)
        call = mock_request.call_args_list[0][0][0]
        self.assertEqual(call.tag, node.tag)
        self.assertEqual(call.attrib, node.attrib)

    @patch('ncclient.transport.SSHSession')
    @patch('ncclient.operations.third_party.juniper.rpc.RPC._request')
    def test_compare_conf(self, mock_request, mock_session):
        device_handler = manager.make_device_handler({'name': 'junos'})
        session = ncclient.transport.SSHSession(device_handler)
        obj = CompareConfiguration(
            session,
            device_handler,
            raise_mode=RaiseMode.ALL)
        obj.request(rollback=2)
        node = new_ele(
            'get-configuration', {'compare': 'rollback', 'rollback': str(2), 'format': 'text'})
        call = mock_request.call_args_list[0][0][0]
        self.assertEqual(call.tag, node.tag)
        self.assertEqual(call.attrib, node.attrib)

    @patch('ncclient.transport.SSHSession')
    @patch('ncclient.operations.third_party.juniper.rpc.RPC._request')
    def test_execute_rpc(self, mock_request, mock_session):
        device_handler = manager.make_device_handler({'name': 'junos'})
        session = ncclient.transport.SSHSession(device_handler)
        obj = ExecuteRpc(session, device_handler, raise_mode=RaiseMode.ALL)
        rpc = new_ele('get-software-information')
        obj.request(rpc)
        mock_request.assert_called_once_with(rpc)

    @patch('ncclient.transport.SSHSession')
    @patch('ncclient.operations.third_party.juniper.rpc.RPC._request')
    def test_execute_rpc_str(self, mock_request, mock_session):
        device_handler = manager.make_device_handler({'name': 'junos'})
        session = ncclient.transport.SSHSession(device_handler)
        obj = ExecuteRpc(session, device_handler, raise_mode=RaiseMode.ALL)
        rpc = '<get-software-information/>'
        obj.request(rpc)
        self.assertEqual(True, isinstance(rpc, str))

    @patch('ncclient.transport.SSHSession')
    @patch('ncclient.operations.third_party.juniper.rpc.RPC._request')
    def test_reboot(self, mock_request, mock_session):
        device_handler = manager.make_device_handler({'name': 'junos'})
        session = ncclient.transport.SSHSession(device_handler)
        obj = Reboot(session, device_handler, raise_mode=RaiseMode.ALL)
        obj.request()
        node = new_ele('request-reboot')
        call = mock_request.call_args_list[0][0][0]
        self.assertEqual(call.tag, node.tag)

    @patch('ncclient.transport.SSHSession')
    @patch('ncclient.operations.third_party.juniper.rpc.RPC._request')
    def test_halt(self, mock_request, mock_session):
        device_handler = manager.make_device_handler({'name': 'junos'})
        session = ncclient.transport.SSHSession(device_handler)
        obj = Halt(session, device_handler, raise_mode=RaiseMode.ALL)
        obj.request()
        node = new_ele('request-halt')
        call = mock_request.call_args_list[0][0][0]
        self.assertEqual(call.tag, node.tag)

    @patch('ncclient.transport.SSHSession')
    @patch('ncclient.operations.third_party.juniper.rpc.RPC._request')
    @patch('ncclient.operations.third_party.juniper.rpc.RPC._assert')
    def test_commit_confirmed(self, mock_assert, mock_request, mock_session):
        # mock_session.server_capabilities.return_value = [':candidate']
        device_handler = manager.make_device_handler({'name': 'junos'})
        session = ncclient.transport.SSHSession(device_handler)
        obj = Commit(session, device_handler, raise_mode=RaiseMode.ALL)
        obj.request(confirmed=True, comment="message", timeout="50")
        node = new_ele_ns("commit-configuration", "")
        sub_ele(node, "confirmed")
        # confirm-timeout need to be provided in minutes for this Junos specific rpc
        sub_ele(node, "confirm-timeout").text = "1"
        sub_ele(node, "log").text = "message"
        xml = ElementTree.tostring(node)
        call = mock_request.call_args_list[0][0][0]
        call = ElementTree.tostring(call)
        self.assertEqual(call, xml)

    @patch('ncclient.transport.SSHSession')
    @patch('ncclient.operations.third_party.juniper.rpc.RPC._request')
    @patch('ncclient.operations.third_party.juniper.rpc.RPC._assert')
    def test_commit(self, mock_assert, mock_request, mock_session):
        # mock_session.server_capabilities.return_value = [':candidate']
        device_handler = manager.make_device_handler({'name': 'junos'})
        session = ncclient.transport.SSHSession(device_handler)
        obj = Commit(session, device_handler, raise_mode=RaiseMode.ALL)
        obj.request()
        node = new_ele_ns("commit-configuration", "")
        xml = ElementTree.tostring(node)
        call = mock_request.call_args_list[0][0][0]
        call = ElementTree.tostring(call)
        self.assertEqual(call, xml)

        obj.request(check=True)
        self.assertEqual(call, xml)

    @patch('ncclient.transport.SSHSession')
    @patch('ncclient.operations.third_party.juniper.rpc.RPC._request')
    @patch('ncclient.operations.third_party.juniper.rpc.RPC._assert')
    def test_commit_at_time(self, mock_assert, mock_request, mock_session):
        # mock_session.server_capabilities.return_value = [':candidate']
        device_handler = manager.make_device_handler({'name': 'junos'})
        session = ncclient.transport.SSHSession(device_handler)
        obj = Commit(session, device_handler, raise_mode=RaiseMode.ALL)
        obj.request(at_time="1111-11-11 00:00:00", synchronize=True)
        node = new_ele_ns("commit-configuration", "")
        sub_ele(node, "at-time").text = "1111-11-11 00:00:00"
        sub_ele(node, "synchronize")
        xml = ElementTree.tostring(node)
        call = mock_request.call_args_list[0][0][0]
        call = ElementTree.tostring(call)
        self.assertEqual(call, xml)

    @patch('ncclient.transport.SSHSession')
    @patch('ncclient.operations.third_party.juniper.rpc.RPC._request')
    @patch('ncclient.operations.third_party.juniper.rpc.RPC._assert')
    def test_commit_confirmed_at_time(
            self, mock_assert, mock_request, mock_session):
        # mock_session.server_capabilities.return_value = [':candidate']
        device_handler = manager.make_device_handler({'name': 'junos'})
        session = ncclient.transport.SSHSession(device_handler)
        obj = Commit(session, device_handler, raise_mode=RaiseMode.ALL)
        self.assertRaises(NCClientError,
            obj.request, at_time="1111-11-11 00:00:00", synchronize=True,
                          confirmed=True)

    @patch('ncclient.transport.SSHSession')
    @patch('ncclient.operations.third_party.juniper.rpc.RPC._request')
    @patch('ncclient.operations.third_party.juniper.rpc.RPC._assert')
    def test_rollback(self, mock_assert, mock_request, mock_session):
        mock_request.return_value = 'junos'
        expected = 'junos'
        # mock_session.server_capabilities.return_value = [':candidate']
        device_handler = manager.make_device_handler({'name': 'junos'})
        session = ncclient.transport.SSHSession(device_handler)
        obj = Rollback(session, device_handler, raise_mode=RaiseMode.ALL)
        actual = obj.request()
        self.assertEqual(expected, actual)

        actual = obj.request(rollback=1)
        self.assertEqual(expected, actual)



import unittest
try:
    from unittest.mock import patch  # Python 3.4 and later
except ImportError:
    from mock import patch
from ncclient import manager
import ncclient.transport
from ncclient.operations.third_party.h3c.rpc import *


class TestRPC(unittest.TestCase):
    def setUp(self):
        self.device_handler = manager.make_device_handler({'name': 'h3c'})
    
    @patch('ncclient.operations.third_party.h3c.rpc.RPC._request')
    def test_getBulk(self, mock_request):
        mock_request.return_value = 'h3c'
        expected = 'h3c'
        session = ncclient.transport.SSHSession(self.device_handler)
        obj = GetBulk(session, self.device_handler)
        actual = obj.request()
        self.assertEqual(expected, actual)

        filter = '<filter xmlns="urn:ietf:params:xml:ns:netconf:base:1.0"></filter>'
        actual = obj.request(filter=filter)
        self.assertEqual(expected, actual)

    @patch('ncclient.operations.third_party.h3c.rpc.RPC._request')
    def test_getBulkConfig(self, mock_request):
        mock_request.return_value = 'h3c'
        expected = 'h3c'
        session = ncclient.transport.SSHSession(self.device_handler)
        obj = GetBulkConfig(session, self.device_handler)
        source = 'devices-name'
        actual = obj.request(source=source)
        self.assertEqual(expected, actual)
        
        filter='<filter xmlns="urn:ietf:params:xml:ns:netconf:base:1.0"></filter>'
        actual = obj.request(source=source, filter=filter)
        self.assertEqual(expected, actual)
        
    @patch('ncclient.operations.third_party.h3c.rpc.RPC._request')
    def test_CLI(self, mock_request):
        mock_request.return_value = 'h3c'
        expected = 'h3c'
        session = ncclient.transport.SSHSession(self.device_handler)
        obj = CLI(session, self.device_handler)
        command = '<get>devices-name</get>'
        actual = obj.request(command=command)
        self.assertEqual(expected, actual)

    @patch('ncclient.operations.third_party.h3c.rpc.RPC._request')
    def test_action(self, mock_request):
        mock_request.return_value = 'h3c'
        expected = 'h3c'
        session = ncclient.transport.SSHSession(self.device_handler)
        obj = Action(session, self.device_handler)
        action = '<get>devices-name</get>'
        actual = obj.request(action=action)
        self.assertEqual(expected, actual)

    @patch('ncclient.operations.third_party.h3c.rpc.RPC._request')
    def test_save(self, mock_request):
        mock_request.return_value = 'h3c'
        expected = 'h3c'
        session = ncclient.transport.SSHSession(self.device_handler)
        obj = Save(session, self.device_handler)
        actual = obj.request()
        self.assertEqual(expected, actual)

        file = 'devices-name'
        actual = obj.request(file=file)
        self.assertEqual(expected, actual)

    @patch('ncclient.operations.third_party.h3c.rpc.RPC._request')
    def test_load(self, mock_request):
        mock_request.return_value = 'h3c'
        expected = 'h3c'
        session = ncclient.transport.SSHSession(self.device_handler)
        obj = Load(session, self.device_handler)
        actual = obj.request()
        self.assertEqual(expected, actual)

        file = 'devices-name'
        actual = obj.request(file=file)
        self.assertEqual(expected, actual)

    @patch('ncclient.operations.third_party.h3c.rpc.RPC._request')
    def test_rollback(self, mock_request):
        mock_request.return_value = 'h3c'
        expected = 'h3c'
        session = ncclient.transport.SSHSession(self.device_handler)
        obj = Rollback(session, self.device_handler)
        actual = obj.request()
        self.assertEqual(expected, actual)

        file = 'devices-name'
        actual = obj.request(file=file)
        self.assertEqual(expected, actual)



import unittest

from ncclient.devices.sros import SrosDeviceHandler, ConfigMode
from ncclient.xml_ import to_ele, to_xml, BASE_NS_1_0
from ncclient.operations.third_party.sros.rpc import MdCliRawCommand, Commit

capabilities = ['urn:ietf:params:netconf:base:1.0',
                'urn:ietf:params:netconf:base:1.1',
                'urn:ietf:params:netconf:capability:writable-running:1.0',
                'urn:ietf:params:netconf:capability:candidate:1.0',
                'urn:ietf:params:netconf:capability:confirmed-commit:1.0',
                'urn:ietf:params:netconf:capability:rollback-on-error:1.0',
                'urn:ietf:params:netconf:capability:startup:1.0',
                'urn:ietf:params:netconf:capability:url:1.0?scheme=http,ftp,file,https,sftp',
                'urn:ietf:params:netconf:capability:validate:1.0',
                'urn:ietf:params:netconf:capability:xpath:1.0',
                'urn:ietf:params:netconf:capability:notification:1.0',
                'urn:ietf:params:netconf:capability:interleave:1.0',
                'urn:ietf:params:netconf:capability:with-defaults:1.0',
                'urn:ietf:params:xml:ns:netconf:base:1.0',
                'urn:ietf:params:xml:ns:yang:1',
                'urn:ietf:params:netconf:capability:confirmed-commit:1.1',
                'urn:ietf:params:netconf:capability:validate:1.1']

xml = """<?xml version="1.0" encoding="UTF-8"?><rpc-reply xmlns:nc="urn:ietf:params:xml:ns:netconf:base:1.0" xmlns="urn:ietf:params:xml:ns:netconf:base:1.0" message-id="urn:uuid:4db8afe2-fbd9-4c13-abaf-f39ef678f63c">
<results xmlns="urn:nokia.com:sros:ns:yang:sr:oper-global">
<md-cli-output-block>TiMOS-B-20.10.B1-5 both/x86_64 Nokia 7750 SR Copyright (c) 2000-2020 Nokia.

All rights reserved. All use subject to applicable license agreements.

Built on Fri Oct 2 18:11:20 PDT 2020 by builder in /builds/c/2010B/B1-5/panos/main/sros

</md-cli-output-block>
</results>
</rpc-reply>"""

class TestSrosDevice(unittest.TestCase):

    def setUp(self):
        self.device_handler = SrosDeviceHandler({'name': 'sros'})

    def test_add_additional_operations(self):
        expected = {
            'md_cli_raw_command': MdCliRawCommand,
            'commit': Commit,
        }
        self.assertDictEqual(expected, self.device_handler.add_additional_operations())

    def test_transform_reply(self):
        expected = xml
        actual = self.device_handler.transform_reply()
        ele = to_ele(xml)
        self.assertEqual(expected, to_xml(actual(ele)))

    def test_get_capabilities_without_config_mode(self):
        """Test capabilities without 'config_mode' set"""
        self.assertListEqual(capabilities, self.device_handler.get_capabilities())

    def test_get_capabilities_with_config_mode_private(self):
        """Test capabilities when 'config_mode' is set to 'private'"""
        device_handler = SrosDeviceHandler({'name': 'sros', 'config_mode': ConfigMode.PRIVATE})
        expected_capabilities = capabilities + ['urn:nokia.com:nc:pc']
        self.assertListEqual(expected_capabilities, device_handler.get_capabilities())

    def test_get_xml_base_namespace_dict(self):
        expected = {None: BASE_NS_1_0}
        self.assertDictEqual(expected, self.device_handler.get_xml_base_namespace_dict())

    def test_get_xml_extra_prefix_kwargs(self):
        expected = dict()
        expected['nsmap'] = self.device_handler.get_xml_base_namespace_dict()
        self.assertDictEqual(expected, self.device_handler.get_xml_extra_prefix_kwargs())

    def test_perform_qualify_check(self):
        self.assertFalse(self.device_handler.perform_qualify_check())



#!/usr/bin/env python
import logging

from ncclient import manager
from ncclient.xml_ import *


def connect(host, port, user, password):
    conn = manager.connect(host=host,
                           port=port,
                           username=user,
                           password=password,
                           timeout=60,
                           device_params={'name': 'junos'},
                           hostkey_verify=False)

    rpc = """
    <get-chassis-inventory>
        <detail/>
    </get-chassis-inventory>"""

    result = conn.rpc(rpc)
    logging.info('Chassis serial-number: %s', result.xpath('//chassis-inventory/chassis/serial-number')[0].text)


if __name__ == '__main__':
    LOG_FORMAT = '%(asctime)s %(levelname)s %(filename)s:%(lineno)d %(message)s'
    logging.basicConfig(stream=sys.stdout, level=logging.INFO, format=LOG_FORMAT)

    connect('router', 830, 'netconf', 'juniper!')



# Copyright 2009 Shikhar Bhushan
# Copyright 201[2-5] Leonidas Poulopoulos (@leopoul)
# Copyright 2013 Ebben Aries
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup, find_packages
from distutils.command.install import install as _install

import sys
import codecs
import versioneer

__author__ = "Shikhar Bhushan, Leonidas Poulopoulos, Ebben Aries, Einar Nilsen-Nygaard"
__author_email__ = "shikhar@schmizz.net, lpoulopoulos@verisign.com, exa@dscp.org, einarnn@gmail.com"
__licence__ = "Apache 2.0"

if sys.version_info.major == 2 or sys.version_info.minor < 5:
    print ("Sorry, Python < 3.5 is not supported")
    exit()

#parse requirements
req_lines = [line.strip() for line in open("requirements.txt").readlines()]
install_reqs = list(filter(None, req_lines))

test_req_lines = [line.strip() for line in open("test-requirements.txt").readlines()]
test_reqs = list(filter(None, test_req_lines))

with codecs.open('README.rst', 'r', encoding='utf8') as file:
    long_description = file.read()


setup(name='ncclient',
      version=versioneer.get_version(),
      cmdclass=versioneer.get_cmdclass(),
      description="Python library for NETCONF clients",
      long_description = long_description,
      author=__author__,
      author_email=__author_email__,
      url="https://github.com/ncclient/ncclient",
      packages=find_packages(exclude=['test', 'test.*']),
      install_requires=install_reqs,
      tests_require=test_reqs,
      license=__licence__,
      platforms=["Posix; OS X; Windows"],
      keywords=['NETCONF', 'NETCONF Python client', 'Juniper Optimization', 'Cisco NXOS Optimization'],
      python_requires='>=3.5',
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Topic :: System :: Networking',
          'Intended Audience :: Developers',
          'Operating System :: OS Independent',
          'Topic :: Software Development :: Libraries :: Python Modules',
      ])










# supported devices config, add new device (eg: 'device name':'device label').
supported_devices_cfg = {
    "alu": "Alcatel Lucent",
    "ciena": "Ciena",
    "csr": "Cisco CSR1000v",
    "h3c": "H3C",
    "hpcomware": "HP Comware",
    "huawei": "Huawei",
    "huaweiyang": "Huawei",
    "iosxe": "Cisco IOS XE",
    "iosxr": "Cisco IOS XR",
    "junos": "Juniper",
    "nexus": "Cisco Nexus",
    "sros": "Nokia SR OS",
    "default": "Server or anything not in above",
}


def get_supported_devices():
    return tuple(supported_devices_cfg.keys())


def get_supported_device_labels():
    return supported_devices_cfg



import unittest
from ncclient.devices.nexus import *


capabilities = ['urn:ietf:params:xml:ns:netconf:base:1.0',
                'urn:ietf:params:netconf:base:1.1',
                'urn:ietf:params:netconf:capability:writable-running:1.0',
                'urn:ietf:params:netconf:capability:candidate:1.0',
                'urn:ietf:params:netconf:capability:confirmed-commit:1.0',
                'urn:ietf:params:netconf:capability:rollback-on-error:1.0',
                'urn:ietf:params:netconf:capability:startup:1.0',
                'urn:ietf:params:netconf:capability:url:1.0?scheme=http,ftp,file,https,sftp',
                'urn:ietf:params:netconf:capability:validate:1.0',
                'urn:ietf:params:netconf:capability:xpath:1.0',
                'urn:ietf:params:netconf:capability:notification:1.0',
                'urn:ietf:params:netconf:capability:interleave:1.0',
                'urn:ietf:params:netconf:capability:with-defaults:1.0']


class TestNexusDevice(unittest.TestCase):

    def setUp(self):
        self.obj = NexusDeviceHandler({'name': 'nexus'})

    def test_add_additional_operations(self):
        expected = {'exec_command': ExecCommand}
        self.assertDictEqual(expected, self.obj.add_additional_operations())

    def test_get_capabilities(self):
        self.assertListEqual(capabilities, self.obj.get_capabilities())

    def test_get_xml_base_namespace_dict(self):
        expected = {None: BASE_NS_1_0}
        self.assertDictEqual(expected, self.obj.get_xml_base_namespace_dict())

    def test_get_xml_extra_prefix_kwargs(self):
        temp = {
            'nxos': 'http://www.cisco.com/nxos:1.0',
            'if': 'http://www.cisco.com/nxos:1.0:if_manager',
            'nfcli': 'http://www.cisco.com/nxos:1.0:nfcli',
            'vlan_mgr_cli': 'http://www.cisco.com/nxos:1.0:vlan_mgr_cli'
        }
        temp.update(self.obj.get_xml_base_namespace_dict())
        expected = dict()
        expected['nsmap'] = temp
        self.assertDictEqual(expected, self.obj.get_xml_extra_prefix_kwargs())

    def test_get_ssh_subsystem_names(self):
        expected = ['netconf', 'xmlagent']
        self.assertListEqual(expected, self.obj.get_ssh_subsystem_names())

        expected.insert(0, 'ncclient')
        self.obj = NexusDeviceHandler({'ssh_subsystem_name': 'ncclient'}) 
        self.assertListEqual(expected, self.obj.get_ssh_subsystem_names())



import unittest
from ncclient.devices.alu import *
from ncclient.xml_ import *
import re


xml = """<rpc-reply xmlns:junos="http://xml.alu.net/alu/12.1x46/alu">
<routing-engin>
<name>reX</name>
<commit-success/>
<!-- This is a comment -->
</routing-engin>
<ok/>
</rpc-reply>"""


class TestAluDevice(unittest.TestCase):
    
    def setUp(self):
        self.obj = AluDeviceHandler({'name': 'alu'})

    def test_remove_namespaces(self):
        xmlObj = to_ele(xml)
        expected = re.sub(r'<rpc-reply xmlns:junos="http://xml.alu.net/alu/12.1x46/alu">',
                          r'<?xml version="1.0" encoding="UTF-8"?><rpc-reply>', xml)
        self.assertEqual(expected, to_xml(remove_namespaces(xmlObj)))

    def test_get_capabilities(self):
        expected = ["urn:ietf:params:netconf:base:1.0", ]
        self.assertListEqual(expected, self.obj.get_capabilities())

    def test_get_xml_base_namespace_dict(self):
        expected = {None: BASE_NS_1_0}
        self.assertDictEqual(expected, self.obj.get_xml_base_namespace_dict())

    def test_get_xml_extra_prefix_kwargs(self):
        expected = dict()
        expected["nsmap"] = self.obj.get_xml_base_namespace_dict()
        self.assertDictEqual(expected, self.obj.get_xml_extra_prefix_kwargs())

    def test_add_additional_operations(self):
        expected=dict()
        expected["get_configuration"] = GetConfiguration
        expected["show_cli"] = ShowCLI
        expected["load_configuration"] = LoadConfiguration
        self.assertDictEqual(expected, self.obj.add_additional_operations())
    
    def test_transform_reply(self):
        expected = re.sub(r'<rpc-reply xmlns:junos="http://xml.alu.net/alu/12.1x46/alu">',
                          r'<?xml version="1.0" encoding="UTF-8"?><rpc-reply>', xml)
        actual = self.obj.transform_reply()
        xmlObj = to_ele(xml)
        self.assertEqual(expected, to_xml(actual(xmlObj)))



#!/usr/bin/env python
import sys

import logging

from ncclient import manager


def connect(host, port, user, password):
    conn = manager.connect(host=host,
                           port=port,
                           username=user,
                           password=password,
                           timeout=60,
                           device_params={'name': 'junos'},
                           hostkey_verify=False)

    conn.lock()

    # configuration as a string
    load_config_result = conn.load_configuration(action='set', config='set system host-name foo')
    logging.info(load_config_result)

    # configuration as a list
    location = []
    location.append('set system location building "Main Campus, C"')
    location.append('set system location floor 15')
    location.append('set system location rack 1117')

    load_config_result = conn.load_configuration(action='set', config=location)
    logging.info(load_config_result)

    validate_result = conn.validate()
    logging.info(validate_result)

    compare_config_result = conn.compare_configuration()
    logging.info(compare_config_result)

    conn.commit()
    conn.unlock()
    conn.close_session()


if __name__ == '__main__':
    LOG_FORMAT = '%(asctime)s %(levelname)s %(filename)s:%(lineno)d %(message)s'
    logging.basicConfig(stream=sys.stdout, level=logging.INFO, format=LOG_FORMAT)

    connect('router', '22', 'netconf', 'juniper!')



from ncclient.operations.lock import *
import unittest
try:
    from unittest.mock import patch  # Python 3.4 and later
except ImportError:
    from mock import patch
from ncclient import manager
import ncclient.manager
import ncclient.transport
from ncclient.xml_ import *
from ncclient.operations import RaiseMode
from xml.etree import ElementTree


class TestLock(unittest.TestCase):

    def setUp(self):
        self.device_handler = manager.make_device_handler({'name': 'junos'})

    @patch('ncclient.transport.SSHSession')
    @patch('ncclient.operations.RPC._request')
    def test_lock_default_param(self, mock_request, mock_session):
        session = ncclient.transport.SSHSession(self.device_handler)
        obj = Lock(session, self.device_handler, raise_mode=RaiseMode.ALL)
        obj.request()
        node = new_ele("lock")
        sub_ele(sub_ele(node, "target"), "candidate")
        xml = ElementTree.tostring(node)
        call = mock_request.call_args_list[0][0][0]
        call = ElementTree.tostring(call)
        self.assertEqual(call, xml)

    @patch('ncclient.transport.SSHSession')
    @patch('ncclient.operations.RPC._request')
    def test_lock(self, mock_request, mock_session):
        session = ncclient.transport.SSHSession(self.device_handler)
        obj = Lock(session, self.device_handler, raise_mode=RaiseMode.ALL)
        obj.request(target="running")
        node = new_ele("lock")
        sub_ele(sub_ele(node, "target"), "running")
        xml = ElementTree.tostring(node)
        call = mock_request.call_args_list[0][0][0]
        call = ElementTree.tostring(call)
        self.assertEqual(call, xml)

    @patch('ncclient.transport.SSHSession')
    @patch('ncclient.operations.RPC._request')
    def test_unlock_default_param(self, mock_request, mock_session):
        session = ncclient.transport.SSHSession(self.device_handler)
        obj = Unlock(session, self.device_handler, raise_mode=RaiseMode.ALL)
        obj.request()
        node = new_ele("unlock")
        sub_ele(sub_ele(node, "target"), "candidate")
        xml = ElementTree.tostring(node)
        call = mock_request.call_args_list[0][0][0]
        call = ElementTree.tostring(call)
        self.assertEqual(call, xml)

    @patch('ncclient.transport.SSHSession')
    @patch('ncclient.operations.RPC._request')
    def test_unlock(self, mock_request, mock_session):
        session = ncclient.transport.SSHSession(self.device_handler)
        obj = Unlock(session, self.device_handler, raise_mode=RaiseMode.ALL)
        obj.request(target="running")
        node = new_ele("unlock")
        sub_ele(sub_ele(node, "target"), "running")
        xml = ElementTree.tostring(node)
        call = mock_request.call_args_list[0][0][0]
        call = ElementTree.tostring(call)
        self.assertEqual(call, xml)

    @patch('ncclient.transport.SSHSession')
    @patch('ncclient.operations.RPC._request')
    def test_lock_context_enter(self, mock_request, mock_session):
        session = ncclient.transport.SSHSession(self.device_handler)
        obj = LockContext(session, self.device_handler, "candidate")
        self.assertEqual(obj.__enter__(), obj)
        node = new_ele("lock")
        sub_ele(sub_ele(node, "target"), "candidate")
        xml = ElementTree.tostring(node)
        call = mock_request.call_args_list[0][0][0]
        call = ElementTree.tostring(call)
        self.assertEqual(call, xml)

    @patch('ncclient.transport.SSHSession')
    @patch('ncclient.operations.RPC._request')
    def test_lock_context_exit(self, mock_request, mock_session):
        session = ncclient.transport.SSHSession(self.device_handler)
        obj = LockContext(session, self.device_handler, "running")
        self.assertFalse(obj.__exit__())
        node = new_ele("unlock")
        sub_ele(sub_ele(node, "target"), "running")
        xml = ElementTree.tostring(node)
        call = mock_request.call_args_list[0][0][0]
        call = ElementTree.tostring(call)
        self.assertEqual(call, xml)



import unittest
from ncclient.devices.iosxr import *


class TestIosxrDevice(unittest.TestCase):
    
    def setUp(self):
        self.obj = IosxrDeviceHandler({'name': 'iosxe'})

    def test_add_additional_ssh_connect_params(self):
        expected = dict()
        expected["unknown_host_cb"] = iosxr_unknown_host_cb
        actual = dict()
        self.obj.add_additional_ssh_connect_params(actual)
        self.assertDictEqual(expected, actual)

    def test_perform_qualify_check(self):
        self.assertFalse(self.obj.perform_qualify_check())

    def test_csr_unknown_host_cb(self):
        self.assertTrue(iosxr_unknown_host_cb('host', 'fingerprint'))
